"""
Tests for the .comm.labjack_ljm subpackage.
"""

import logging

import pytest

from hvl_ccb.comm import (
    LJMCommunication,
    LJMCommunicationConfig,
    LJMCommunicationError,
)

logging.disable(logging.ERROR)


@pytest.fixture(scope='module')
def testconfig():
    return {
        'identifier': '-2'
    }


@pytest.fixture
def com(testconfig):
    with LJMCommunication(testconfig) as com:
        yield com


def test_labjack_config_cleaning():
    with pytest.raises(ValueError):
        LJMCommunicationConfig(device_type='T5')

    with pytest.raises(ValueError):
        LJMCommunicationConfig(connection_type='Gardena')


def test_labjack_config_enums():
    config = LJMCommunicationConfig(device_type=LJMCommunicationConfig.DeviceType.T4)
    assert str(config.device_type) == "T4"

    config = LJMCommunicationConfig(
        connection_type=LJMCommunicationConfig.ConnectionType.TCP)
    assert str(config.connection_type) == "TCP"


def test_open(testconfig):
    com = LJMCommunication(testconfig)
    com.open()
    com.open()

    failing_config = dict(testconfig)
    failing_config['identifier'] = 'this_is_an_invalid_identifier'
    com = LJMCommunication(failing_config)
    with pytest.raises(LJMCommunicationError):
        com.open()


def test_read_name(com):
    with pytest.raises(LJMCommunicationError):
        com.read_name("SERIAL_NUMBER")

    with pytest.raises(LJMCommunicationError):
        com.read_name("SERIAL_NUMBER", "SERIAL_NUMBER")

    with pytest.raises(TypeError):
        com.read_name(123)


def test_write_name(com):
    with pytest.raises(LJMCommunicationError):
        com.write_name("test", 0)

    with pytest.raises(LJMCommunicationError):
        com.write_names(["test1", "test2"], [1, 2])

    with pytest.raises(TypeError):
        com.write_name(123, 456)


def test_write_address(com):
    with pytest.raises(NotImplementedError):
        com.write_address(1234, 5678)



# -*- coding: utf-8 -*-

"""Top-level package for HVL Common Code Base."""

__author__ = """Mikołaj Rybiński, David Graber, Henrik Menne"""
__email__ = 'mikolaj.rybinski@id.ethz.ch, graber@eeh.ee.ethz.ch, ' \
            'henrik.menne@eeh.ee.ethz.ch'
__version__ = '0.3.0'

from . import comm  # noqa: F401
from . import dev  # noqa: F401
from .configuration import ConfigurationMixin, configdataclass  # noqa: F401
from .experiment_manager import (  # noqa: F401
    ExperimentManager,
    ExperimentStatus,
    ExperimentError
)



"""
Tests for the Device SE ILS2T stepper motor.
"""

from queue import Queue
from typing import Any

import pytest

from hvl_ccb.comm import ModbusTcpCommunication, ModbusTcpConnectionFailedException
from hvl_ccb.dev import (
    ILS2T,
    ILS2TConfig,
    IoScanningModeValueError,
    ScalingFactorValueError,
    ILS2TException,
)


class MaskedModbusTcpCommunication(ModbusTcpCommunication):
    def __init__(self, configuration):
        super().__init__(configuration)

        self.simulate_running = True
        self.simulate_enabled = True
        self.simulate_error = False

        self._write_buffer = Queue()

    def open(self):
        pass

    def close(self):
        pass

    def write_registers(self, address: int, values: Any):
        pair = (address, values)
        self._write_buffer.put(pair)

    def get_written(self):
        return self._write_buffer.get() if not self._write_buffer.empty() else None

    def read_holding_registers(self, address: int, count: int):
        if address == ILS2T.RegAddr.IO_SCANNING and count == 8:
            # read IOscanning registers
            if self.simulate_running and self.simulate_enabled:
                # {'mode': 3, 'action': 2, 'ref_16': 1500, 'ref_32': 16000000,
                # 'state': 6, 'halt': False, 'turning_positive': True,
                # 'turning_negative': False}
                return [547, 1500, 244, 9216, 6, 816, 16512, 0]
            elif not self.simulate_running and self.simulate_enabled:
                # {'mode': 3, 'action': 2, 'ref_16': 1500, 'ref_32': 0, 'state': 6,
                # 'fault': False, 'warn': False, 'halt': False, 'motion_zero': True,
                # 'turning_positive': False, 'turning_negative': False}
                return [675, 1500, 0, 0, 24582, 49968, 2112, 0]
            else:
                # {'mode': 3, 'action': 2, 'ref_16': 1500, 'ref_32': 0, 'state': 4,
                # 'halt': True, 'turning_positive': False, 'turning_negative': False}
                return [291, 1500, 0, 0, 57348, 816, 2112, 0]

        return [i for i in range(count)]

    def read_input_registers(self, address: int, count: int):
        if address == ILS2T.RegAddr.TEMP and count == 2:
            # read out temperature
            return [0, 35]

        if address == ILS2T.RegAddr.POSITION and count == 2:
            # read out position
            return [1, 2_599]

        if address == ILS2T.RegAddr.VOLT and count == 2:
            # read out DC voltage
            return [0, 480]  # 48.0 V

        if address == ILS2T.RegAddr.FLT_INFO and count == 22:
            return [0] * count if self.simulate_error else [i for i in range(count)]

        return [i for i in range(count)]


@pytest.fixture(scope='module')
def com_config():
    return {
        'host': '127.0.0.1',
        'unit': 0,
        'port': 502,
    }


@pytest.fixture
def testdev(com_config):
    modbus_tcp = MaskedModbusTcpCommunication(com_config)
    motor = ILS2T(modbus_tcp)
    assert motor is not None
    motor.start()
    assert modbus_tcp.get_written() == (motor.RegAddr.ACCESS_ENABLE.value, [0, 1])
    assert modbus_tcp.get_written() == (
        motor.RegAddr.RAMP_N_MAX.value, [0, ILS2TConfig.rpm_max_init])
    yield motor
    motor.stop()


def test_clean_ioscanning_mode_values(testdev: ILS2T):
    default_values = testdev.DEFAULT_IO_SCANNING_CONTROL_VALUES

    # default values need to work
    assert testdev._clean_ioscanning_mode_values(default_values) == default_values

    # missing key that gets loaded from defaults
    other_values = dict(default_values)
    other_values.pop('mode')
    assert testdev._clean_ioscanning_mode_values(other_values) == default_values

    # additional key that is not needed
    other_values = dict(default_values)
    other_values['some_other_key'] = None
    with pytest.raises(ValueError):
        testdev._clean_ioscanning_mode_values(other_values)

    # invalid mode
    other_values = dict(default_values)
    other_values['mode'] = -1
    with pytest.raises(IoScanningModeValueError):
        testdev._clean_ioscanning_mode_values(other_values)

    # with mode 1: JOG
    # ----------------
    other_values = dict(default_values)
    other_values['mode'] = 1

    # action has to be 0
    other_values['action'] = -1
    with pytest.raises(IoScanningModeValueError):
        testdev._clean_ioscanning_mode_values(other_values)

    other_values['action'] = 0
    # action is 0, but ref_16 is still 1500
    with pytest.raises(IoScanningModeValueError):
        testdev._clean_ioscanning_mode_values(other_values)

    # one of the valid values for ref_16 in JOG mode
    other_values['ref_16'] = 1
    assert testdev._clean_ioscanning_mode_values(other_values) == other_values

    # action and ref_16 valid, but ref_32 not `0`
    other_values['ref_32'] = 1
    with pytest.raises(IoScanningModeValueError):
        testdev._clean_ioscanning_mode_values(other_values)

    # with mode 3: Point to Point
    # ---------------------------
    other_values = dict(default_values)
    other_values['mode'] = 3

    # action has to be 1, 2 or 3
    other_values['action'] = 4
    with pytest.raises(IoScanningModeValueError):
        testdev._clean_ioscanning_mode_values(other_values)

    other_values['action'] = 1

    # too high RPM
    other_values = dict(default_values)
    other_values['ref_16'] = 4000  # too high RPM
    with pytest.raises(IoScanningModeValueError):
        testdev._clean_ioscanning_mode_values(other_values)

    other_values['ref_16'] = 1500
    # needs to work
    assert testdev._clean_ioscanning_mode_values(other_values) == other_values


def test_32bit_decode():
    assert ILS2T._decode_32bit([2**15, 1], signed=False) == 2_147_483_649
    assert ILS2T._decode_32bit([2**15, 1], signed=True) == -2_147_483_647


def test_toggle(testdev: ILS2T):
    actual = testdev._mode_toggle_mt
    testdev._toggle()
    assert actual is not testdev._mode_toggle_mt


def test_start_stop_failing(com_config):
    test_device = ILS2T(com_config)

    with pytest.raises(ModbusTcpConnectionFailedException):
        test_device.start()

    with pytest.raises(ModbusTcpConnectionFailedException):
        test_device.stop()


def test_get_dc_volt(testdev: ILS2T):
    assert testdev.get_dc_volt() == 48.0


def test_get_temperature(testdev: ILS2T):
    assert testdev.get_temperature() == 35


def test_get_position(testdev: ILS2T):
    assert testdev.get_position() == 68_135


def test_enable(testdev: ILS2T):
    testdev.enable()
    assert (
        testdev.com.get_written() ==
        (testdev.RegAddr.IO_SCANNING.value,
            testdev._generate_control_registers(enable_driver_en=1))
    )


def test_disable(testdev: ILS2T):
    testdev.com.simulate_running = False
    testdev.disable()
    assert (
        testdev.com.get_written() ==
        (testdev.RegAddr.IO_SCANNING.value, testdev._generate_control_registers(
            disable_driver_di=1))
    )

    testdev.com.simulate_running = True
    testdev.disable()
    assert testdev.com.get_written() is None


def test_relative_step(testdev: ILS2T):
    testdev.com.simulate_enabled = False
    testdev.com.simulate_running = False
    testdev.com.simulate_error = False

    testdev.relative_step(16_000)

    with pytest.raises(IoScanningModeValueError):
        testdev.relative_step(5_000_000_000)

    testdev.relative_step_and_wait(50_000)

    testdev.com.simulate_error = True
    testdev.relative_step_and_wait(10000)


def test_absolute_position(testdev: ILS2T):
    testdev.com.simulate_enabled = False
    testdev.com.simulate_running = False
    testdev.com.simulate_error = False

    testdev.absolute_position(0)

    with pytest.raises(IoScanningModeValueError):
        testdev.absolute_position(5_000_000_000)

    testdev.com.simulate_running = False

    testdev.absolute_position_and_wait(100_000)

    testdev.com.simulate_error = True
    testdev.absolute_position_and_wait(10000)


def test_set_max_rpm(testdev: ILS2T):
    testdev.set_max_rpm(500)
    assert (
        testdev.com.get_written() == (
            testdev.RegAddr.RAMP_N_MAX.value, [0, 500])
    )

    with pytest.raises(ILS2TException):
        testdev.set_max_rpm(-100)
    assert testdev.com.get_written() is None


def test_set_ramp_type(testdev: ILS2T):
    testdev.set_ramp_type(0)
    assert (
        testdev.com.get_written() == (
            testdev.RegAddr.RAMP_TYPE.value, [0, 0])
    )


def test_user_steps(testdev: ILS2T):
    testdev.user_steps(10, 1)
    assert (
        testdev.com.get_written() == (
            testdev.RegAddr.SCALE.value, [0, 10, 0, 1])
    )

    with pytest.raises(ScalingFactorValueError):
        testdev.user_steps(5.5, 3)
    assert testdev.com.get_written() is None

    with pytest.raises(ScalingFactorValueError):
        testdev.user_steps(5, 3.75)
    assert testdev.com.get_written() is None


def test_quickstop(testdev: ILS2T):
    testdev.com.simulate_running = True
    testdev.quickstop()
    testdev.reset_error()


def test_get_status(testdev: ILS2T):
    testdev.get_status()


def test_jog_run(testdev: ILS2T):
    for running in (True, False):
        for enabled in (True, False):
            testdev.com.simulate_enabled = enabled
            testdev.com.simulate_running = running
            testdev.jog_run(direction=True)
            testdev.jog_run(direction=False)
            testdev.jog_run(direction=True, fast=True)
            testdev.jog_run(direction=False, fast=True)
            testdev.jog_stop()


def test_set_jog_speed(testdev: ILS2T):
    testdev.set_jog_speed(1, 1000)
    assert (
        testdev.com.get_written() == (
            testdev.RegAddr.JOGN_SLOW.value, [0, 1])
    )
    assert (
        testdev.com.get_written() == (
            testdev.RegAddr.JOGN_FAST.value, [0, 1000])
    )


def test_set_max_acceleration(testdev: ILS2T):
    testdev.set_max_acceleration(5000)
    assert (
        testdev.com.get_written() == (
            testdev.RegAddr.RAMP_ACC.value, [0, 5000])
    )


def test_set_max_deceleration(testdev: ILS2T):
    testdev.set_max_deceleration(5000)
    assert (
        testdev.com.get_written() == (
            testdev.RegAddr.RAMP_DECEL.value, [0, 5000])
    )



"""
Facilities providing classes for handling configuration for communication protocols
and devices.
"""

import dataclasses
import json
from abc import ABC, abstractmethod
from typing import Dict, Sequence

from .utils.typing import is_generic, check_generic_type


def _has_default_value(f: dataclasses.Field):
    return not isinstance(f.default, dataclasses._MISSING_TYPE)


# Hooks of configdataclass
def _clean_values(self):
    """
    Cleans and enforces configuration values. Does nothing by default, but may be
    overridden to add custom configuration value checks.
    """
    pass


_configclass_hooks = {
    'clean_values': _clean_values,
}


# Methods of configdataclass
def ___post_init__(self):
    self._check_types()
    self.clean_values()


def _force_value(self, fieldname, value):
    """
    Forces a value to a dataclass field despite the class being frozen.

    :param fieldname: name of the field
    :param value: value to assign
    """
    object.__setattr__(self, fieldname, value)


@classmethod
def _keys(cls) -> Sequence[str]:
    """
    Returns a list of all configdataclass fields key-names.

    :return: a list of strings containing all keys.
    """
    return [f.name for f in dataclasses.fields(cls)]


@classmethod
def _required_keys(cls) -> Sequence[str]:
    """
    Returns a list of all configdataclass fields, that have no default value assigned
    and need to be specified on instantiation.

    :return: a list of strings containing all required keys.
    """
    return [
        f.name for f in dataclasses.fields(cls)
        if not _has_default_value(f)
    ]


@classmethod
def _optional_defaults(cls) -> Dict[str, object]:
    """
    Returns a list of all configdataclass fields, that have a default value assigned
    and may be optionally specified on instantiation.

    :return: a list of strings containing all optional keys.
    """
    return {
        f.name: f.default for f in dataclasses.fields(cls)
        if _has_default_value(f)
    }


def __check_types(self):
    for field in dataclasses.fields(self):
        name = field.name
        value = getattr(self, name)
        type_ = field.type
        if is_generic(type_):
            check_generic_type(value, type_, name=name)
        elif not isinstance(value, type_):
            raise TypeError('Type of field `{}` is `{}` and does not match `{}`.'
                            .format(name, type(value), type_))


_configclass_methods = {
    '__post_init__': ___post_init__,
    'force_value': _force_value,
    'keys': _keys,
    'required_keys': _required_keys,
    'optional_defaults': _optional_defaults,
    '_check_types': __check_types,
}


def configdataclass(direct_decoration=None, frozen=True):
    """
    Decorator to make a class a configdataclass. Types in these dataclasses are
    enforced. Implement a function clean_values(self) to do additional checking on
    value ranges etc.

    It is possible to inherit from a configdataclass and re-decorate it with
    @configdataclass. In a subclass, default values can be added to existing fields.
    Note: adding additional non-default fields is prone to errors, since the order
    has to be respected through the whole chain (first non-default fields, only then
    default-fields).

    :param frozen: defaults to True. False allows to later change configuration values.
        Attention: if configdataclass is not frozen and a value is changed, typing is
        not enforced anymore!
    """
    def decorator(cls):
        for name, method in _configclass_methods.items():
            if name in cls.__dict__:
                raise AttributeError(
                    'configdataclass {!r} cannot define {!r} method'.format(
                        cls.__name__,
                        name,
                    )
                )
            setattr(cls, name, method)
        for name, hook in _configclass_hooks.items():
            if not hasattr(cls, name):
                setattr(cls, name, hook)
        if not hasattr(cls, 'is_configdataclass'):
            setattr(cls, 'is_configdataclass', True)

        return dataclasses.dataclass(cls, frozen=frozen)

    if direct_decoration:
        return decorator(direct_decoration)

    return decorator


class ConfigurationMixin(ABC):
    """
    Mixin providing configuration to a class.
    """

    # omitting type hint of `configuration` on purpose, because type hinting
    # configdataclass is not possible. Union[Dict[str, object], object] resolves to
    # object.
    def __init__(self, configuration) -> None:
        """
        Constructor for the configuration mixin.

        :param configuration: is the configuration provided either as:
        *   a dict with string keys and values, then the default config dataclass
            will be used
        *   a configdataclass object
        *   None, then the config_cls() with no parameters is instantiated
        """

        if not configuration:
            configuration = {}

        if hasattr(configuration, 'is_configdataclass'):
            self._configuration = configuration
        elif isinstance(configuration, Dict):
            default_configdataclass = self.config_cls()
            if not hasattr(default_configdataclass, 'is_configdataclass'):
                raise TypeError('Default configdataclass is not a configdataclass. Is'
                                'the decorator `@configdataclass` applied?')
            self._configuration = default_configdataclass(**configuration)
        else:
            raise TypeError('configuration is not a dictionary or configdataclass.')

    @staticmethod
    @abstractmethod
    def config_cls():
        """
        Return the default configdataclass class.

        :return: a reference to the default configdataclass class
        """
        pass  # pragma: no cover

    @property
    def config(self):
        """
        ConfigDataclass property.

        :return: the configuration
        """

        return self._configuration

    @classmethod
    def from_json(cls, filename: str):
        """
        Instantiate communication protocol using configuration from a JSON file.

        :param filename: Path and filename to the JSON configuration
        """

        configuration = cls._configuration_load_json(filename)
        return cls(configuration)

    def configuration_save_json(self, path: str) -> None:
        """
        Save current configuration as JSON file.

        :param path: path to the JSON file.
        """

        self._configuration_save_json(dataclasses.asdict(self._configuration), path)

    @staticmethod
    def _configuration_load_json(path: str) -> Dict[str, object]:
        """
        Load configuration from JSON file and return Dict. This method is only used
        during construction, if not directly a configuration is given but rather a
        path to a JSON config file.

        :param path: Path to the JSON configuration file.
        :return: Dictionary containing the parameters read from the JSON file.
        """

        with open(path, 'r') as fp:
            return json.load(fp)

    @staticmethod
    def _configuration_save_json(configuration: Dict[str, object], path: str) -> None:
        """
        Store a configuration dict to a JSON file.

        :param configuration: configuration dictionary
        :param path: path to the JSON file.
        """

        with open(path, 'w') as fp:
            json.dump(configuration, fp, indent=4)



"""
Supercube package with implementation for system versions from 2019 on (new concept
with hard-PLC Siemens S7-1500 as CPU).
"""

from .base import (  # noqa: F401
    SupercubeConfiguration,
)
from .typ_a import SupercubeWithFU  # noqa: F401
from .typ_b import SupercubeB  # noqa: F401



"""
Tests for the .dev.mbw973 sub-package.
"""

from queue import Queue

import pytest

from hvl_ccb import comm, dev


@pytest.fixture(scope="module")
def com_config():
    return {
        "port": "loop://?logging=debug",
        "baudrate": 115200,
        "parity": dev.MBW973SerialCommunicationConfig.Parity.NONE,
        "stopbits": dev.MBW973SerialCommunicationConfig.Stopbits.ONE,
        "bytesize": dev.MBW973SerialCommunicationConfig.Bytesize.EIGHTBITS,
        "terminator": b'\r\n',
        "timeout": 3,
    }


@pytest.fixture(scope='module')
def dev_config():
    return {
        'polling_interval': 0.1
    }


@pytest.fixture
def started_mbw973_device(com_config, dev_config):
    serial_port = LoopSerialCommunication(com_config)
    serial_port.open()
    serial_port.put_text('1')
    serial_port.put_text('0')
    with dev.MBW973(serial_port, dev_config) as mbw:
        while serial_port.get_written() is not None:
            pass
        yield serial_port, mbw


class LoopSerialCommunication(comm.SerialCommunication):
    """
    Serial communication for the tests "loop://" port. Masks `write_text` method and
    adds `put_text` method to put actual values for the serial communication protocol to
    read with the `read_text` method.
    """

    def __init__(self, configuration):
        super().__init__(configuration)

        self._write_buffer = Queue()

    def write_text(self, text: str):
        self._write_buffer.put(text)

    def put_text(self, text: str):
        super().write_text(text)

    def get_written(self):
        return self._write_buffer.get() if not self._write_buffer.empty() else None


def test_instantiation(com_config, dev_config):
    mbw = dev.MBW973(com_config)
    assert mbw is not None

    wrong_config = dict(dev_config)
    wrong_config['polling_interval'] = 0
    with pytest.raises(ValueError):
        dev.MBW973(com_config, wrong_config)


def test_start(started_mbw973_device, com_config):
    com, mbw = started_mbw973_device

    # starting again should work
    com.put_text('1')
    com.put_text('0')
    mbw.start()
    assert com.get_written() == 'HumidityTest?'
    assert com.get_written() == 'SF6PurityTest?'

    # assigning new com which is illegal
    com = LoopSerialCommunication(com_config)
    com.ser.port = None
    mbw = dev.MBW973(com)
    with pytest.raises(dev.MBW973Error):
        mbw.start()


def test_is_done(started_mbw973_device):
    serial_port, mbw = started_mbw973_device

    # case 1, only humidity
    mbw.set_measuring_options(humidity=True, sf6_purity=False)
    assert serial_port.get_written() == 'HumidityTest=1'
    assert serial_port.get_written() == 'SF6PurityTest=0'
    serial_port.put_text('0')
    assert mbw.is_done() is False
    assert serial_port.get_written() == 'DoneWithDP?'

    # case 2, only SF6 purity
    mbw.set_measuring_options(humidity=False, sf6_purity=True)
    assert serial_port.get_written() == 'HumidityTest=0'
    assert serial_port.get_written() == 'SF6PurityTest=1'
    serial_port.put_text('0')
    assert mbw.is_done() is False
    assert serial_port.get_written() == 'SF6VolHold?'

    # case 3, both measurements
    mbw.set_measuring_options(humidity=True, sf6_purity=True)
    assert serial_port.get_written() == 'HumidityTest=1'
    assert serial_port.get_written() == 'SF6PurityTest=1'
    serial_port.put_text('1')
    serial_port.put_text('0')
    assert mbw.is_done() is False
    assert serial_port.get_written() == 'DoneWithDP?'
    assert serial_port.get_written() == 'SF6VolHold?'

    # case 3, both measurements, resulting in True
    mbw.set_measuring_options(humidity=True, sf6_purity=True)
    assert serial_port.get_written() == 'HumidityTest=1'
    assert serial_port.get_written() == 'SF6PurityTest=1'
    serial_port.put_text('1')
    serial_port.put_text('1')
    serial_port.put_text('-17.2315')
    serial_port.put_text('-21.5501')
    serial_port.put_text('170.421')
    serial_port.put_text('87.2323')
    serial_port.put_text('35.0001')
    serial_port.put_text('87.53')
    assert mbw.is_done() is True
    assert mbw.last_measurement_values == {
        'frostpoint': -17.2315,
        'frostpoint_ambient': -21.5501,
        'pressure': 170.421,
        'ppmv': 87.2323,
        'ppmw': 35.0001,
        'sf6_vol': 87.53,
    }

    mbw.stop()


def test_mbw973_start_control(started_mbw973_device):
    """
    Tests for the start_control method
    """

    serial_port, mbw = started_mbw973_device

    # control still running
    serial_port.ser.reset_input_buffer()
    serial_port.put_text('1')
    with pytest.raises(dev.MBW973ControlRunningException):
        mbw.start_control()

    assert serial_port.get_written() == 'control?'

    # pump still running
    serial_port.put_text('0')
    serial_port.put_text('1')
    with pytest.raises(dev.MBW973PumpRunningException):
        mbw.start_control()

    assert serial_port.get_written() == 'control?'
    assert serial_port.get_written() == 'Pump.on?'

    # nothing running, can start
    serial_port.put_text('0')  # no control running
    serial_port.put_text('0')  # no pump running
    mbw.start_control()
    assert serial_port.get_written() == 'control?'
    assert serial_port.get_written() == 'Pump.on?'
    assert serial_port.get_written() == 'control=1'

    serial_port.put_text('0')  # not done with dewpoint
    assert mbw.is_done_with_measurements is False
    mbw.polling_timer.t.join()
    assert mbw.is_done_with_measurements is False
    serial_port.put_text('1')  # done with dewpoint
    # dummy measurement values in the readout order
    # note: for Py >= 3.6 dict items are ordered by default (no need to use OrderedDict)
    expected_measurement_values = {
        'frostpoint': 0.1,
        'frostpoint_ambient': 0.2,
        'pressure': 0.3,
        'ppmv': 0.4,
        'ppmw': 0.5,
        'sf6_vol': 0.6,
    }
    for measurement_value in expected_measurement_values.values():
        serial_port.put_text(str(measurement_value))
    mbw.polling_timer.t.join()
    assert mbw.is_done_with_measurements is True
    assert mbw.last_measurement_values == expected_measurement_values


def test_read_measurements(started_mbw973_device):
    com, mbw = started_mbw973_device

    expected_measurement_values = {
        'frostpoint': 0.1,
        'frostpoint_ambient': 0.2,
        'pressure': 0.3,
        'ppmv': 0.4,
        'ppmw': 0.5,
        'sf6_vol': 0.6,
    }
    for measurement_value in expected_measurement_values.values():
        com.put_text(str(measurement_value))

    readout = mbw.read_measurements()

    assert readout == expected_measurement_values

    assert com.get_written() == 'Fp?'
    assert com.get_written() == 'Fp1?'
    assert com.get_written() == 'Px?'
    assert com.get_written() == 'PPMv?'
    assert com.get_written() == 'PPMw?'
    assert com.get_written() == 'SF6Vol?'


def test_set_measuring_options(started_mbw973_device):
    test_com, mbw = started_mbw973_device

    mbw.set_measuring_options(True, True)
    assert test_com.get_written() == 'HumidityTest=1'
    assert test_com.get_written() == 'SF6PurityTest=1'

    mbw.set_measuring_options(True, False)
    assert test_com.get_written() == 'HumidityTest=1'
    assert test_com.get_written() == 'SF6PurityTest=0'

    mbw.set_measuring_options(False, True)
    assert test_com.get_written() == 'HumidityTest=0'
    assert test_com.get_written() == 'SF6PurityTest=1'

    mbw.set_measuring_options(False, False)
    assert test_com.get_written() == 'HumidityTest=0'
    assert test_com.get_written() == 'SF6PurityTest=0'



"""
Tests for the R&S RTO1024 digital storage oscilloscope device class.
"""

import pytest

from hvl_ccb.dev import RTO1024, RTO1024Error, RTO1024VisaCommunicationConfig
from tests.masked_comm import MaskedVisaCommunication


@pytest.fixture(scope='module')
def com_config():
    return {
        'interface_type': RTO1024VisaCommunicationConfig.InterfaceType.TCPIP_INSTR,
        'host': '127.0.0.1',
        'open_timeout': 10,
        'timeout': 50,
    }


@pytest.fixture(scope='module')
def dev_config():
    return {
        'waveforms_path': 'C:\\Data\\DavidGraber\\02_waveforms',
        'settings_path': 'C:\\Data\\DavidGraber\\01_settings',
        'backup_path': 'D:\\backups',
        'command_timeout_seconds': 0.05,
        'spoll_interval': 0.01,
        'spoll_start_delay': 0,
    }


@pytest.fixture()
def testdev(com_config, dev_config):
    com = MaskedVisaCommunication(com_config)
    dev = RTO1024(com, dev_config)
    dev.SPOLL_INTERVAL = 0.01
    dev.SPOLL_START_DELAY = 0
    assert dev is not None
    dev.start()
    assert dev.com.get_written() == '&GTR'
    assert dev.com.get_written() == '*RST'
    assert dev.com.get_written() == '*CLS'
    assert dev.com.get_written() == 'SYST:DISP:UPD ON'
    assert dev.com.get_written() == '*ESE 127'
    yield dev
    dev.stop()


def test_instantiation(com_config, dev_config):
    copy_config = dict(com_config)
    copy_config.pop('interface_type')
    dev = RTO1024(copy_config, dev_config)
    assert dev is not None

    dev_config_dict = dev_config

    assert dev.config.spoll_interval == dev_config_dict['spoll_interval']

    invalid_dev_config = dict(dev_config_dict)
    invalid_dev_config['spoll_interval'] = 0
    with pytest.raises(ValueError):
        RTO1024(com_config, invalid_dev_config)

    assert dev.config.spoll_start_delay == dev_config_dict['spoll_start_delay']

    invalid_dev_config = dict(dev_config_dict)
    invalid_dev_config['spoll_start_delay'] = -1
    with pytest.raises(ValueError):
        RTO1024(com_config, invalid_dev_config)

    assert dev.config.command_timeout_seconds == dev_config_dict[
        'command_timeout_seconds']

    invalid_dev_config = dict(dev_config_dict)
    invalid_dev_config['command_timeout_seconds'] = 0
    with pytest.raises(ValueError):
        RTO1024(com_config, invalid_dev_config)


def test_local_display(testdev: RTO1024):
    testdev.local_display(True)
    assert testdev.com.get_written() == 'SYST:DISP:UPD ON'

    testdev.local_display(False)
    assert testdev.com.get_written() == 'SYST:DISP:UPD OFF'


def test_set_acquire_length(testdev: RTO1024):
    testdev.set_acquire_length(10)
    assert testdev.com.get_written() == 'TIMebase:RANGe 10'


def test_set_reference_point(testdev: RTO1024):
    testdev.set_reference_point(50)
    assert testdev.com.get_written() == 'TIMebase:REFerence 50'


def test_set_repetitions(testdev: RTO1024):
    testdev.set_repetitions(10)
    assert testdev.com.get_written() == 'ACQuire:COUNt 10'


def test_set_channel_state(testdev: RTO1024):
    testdev.set_channel_state(2, True)
    assert testdev.com.get_written() == 'CHANnel2:STATe ON'

    testdev.set_channel_state(3, False)
    assert testdev.com.get_written() == 'CHANnel3:STATe OFF'


def test_set_channel_scale(testdev: RTO1024):
    testdev.set_channel_scale(1, 1e-3)
    assert testdev.com.get_written() == 'CHANnel1:SCALe 0.001'


def test_set_channel_range(testdev: RTO1024):
    testdev.set_channel_range(2, 1.5)
    assert testdev.com.get_written() == 'CHANnel2:RANGe 1.500'


def test_set_channel_position(testdev: RTO1024):
    testdev.set_channel_position(4, 4)
    assert testdev.com.get_written() == 'CHANnel4:POSition 4.00'


def test_set_trigger_source(testdev: RTO1024):
    testdev.set_trigger_source(1)  # to channel 1
    assert testdev.com.get_written() == 'TRIGger1:SOURce CHAN1'

    testdev.set_trigger_source(3, 2)  # to channel 3, B-Event
    assert testdev.com.get_written() == 'TRIGger2:SOURce CHAN3'


def test_set_trigger_level(testdev: RTO1024):
    testdev.set_trigger_level(1, 2e-3)
    assert testdev.com.get_written() == 'TRIGger1:LEVel1 0.002'


def test_set_trigger_mode(testdev: RTO1024):
    testdev.set_trigger_mode(RTO1024.TriggerModes.AUTO)
    assert testdev.com.get_written() == 'TRIGger1:MODE AUTO'

    testdev.set_trigger_mode(RTO1024.TriggerModes.NORMAL)
    assert testdev.com.get_written() == 'TRIGger1:MODE NORMAL'

    testdev.set_trigger_mode(RTO1024.TriggerModes.FREERUN)
    assert testdev.com.get_written() == 'TRIGger1:MODE FREERUN'

    testdev.set_trigger_mode('Auto')
    assert testdev.com.get_written() == 'TRIGger1:MODE AUTO'

    testdev.set_trigger_mode('normal')
    assert testdev.com.get_written() == 'TRIGger1:MODE NORMAL'

    testdev.set_trigger_mode('FREERUN')
    assert testdev.com.get_written() == 'TRIGger1:MODE FREERUN'

    with pytest.raises(RTO1024Error):
        testdev.set_trigger_mode('myunknownmode')

    assert testdev.com.get_written() is None


def test_file_copy(testdev: RTO1024):
    testdev.com.stb = 32
    testdev.file_copy('C:\\Data\\test.txt', 'D:\\test.txt')
    testdev.com.stb = 0

    assert testdev.com.get_written() == '*CLS'
    assert (testdev.com.get_written() ==
            "MMEMory:COPY 'C:\\Data\\test.txt', 'D:\\test.txt'")
    assert testdev.com.get_written() == '*OPC'

    # test timeout
    with pytest.raises(RTO1024Error):
        testdev.file_copy('C:\\Data\\test.txt', 'D:\\test.txt')


def test_backup_waveform(testdev: RTO1024):
    testdev.com.stb = 32
    testdev.backup_waveform('test')
    assert testdev.com.get_written() == '*CLS'
    assert (
        testdev.com.get_written() ==
        "MMEMory:COPY 'C:\\Data\\DavidGraber\\02_waveforms\\test.Wfm.bin', "
        "'D:\\backups\\test.Wfm.bin'"
    )
    assert testdev.com.get_written() == '*OPC'
    assert testdev.com.get_written() == '*CLS'
    assert (
        testdev.com.get_written() ==
        "MMEMory:COPY 'C:\\Data\\DavidGraber\\02_waveforms\\test.bin', "
        "'D:\\backups\\test.bin'"
    )
    assert testdev.com.get_written() == '*OPC'
    testdev.com.stb = 0


def test_list_directory(testdev: RTO1024):
    testdev.com.put_name(
        "MMEMory:CATalog? 'C:/Data'",
        '142422792,147771314176,".,DIR,0","..,DIR,0","DavidGraber,DIR,0","Myriam,DIR,'
        '0","test.Wfm.bin,BIN,142422792"'
    )

    dir_list = testdev.list_directory('C:/Data')
    expected = [
        ['DavidGraber', 'DIR', '0'],
        ['Myriam', 'DIR', '0'],
        ['test.Wfm.bin', 'BIN', '142422792']
    ]

    assert dir_list == expected


def test_save_waveform_history(testdev: RTO1024):
    testdev.com.put_name(
        "MMEMory:CATalog? '{}'".format(testdev.config.waveforms_path),
        '142422792,147771314176,".,DIR,0","..,DIR,0","DavidGraber,DIR,0","Myriam,DIR,'
        '0","test.Wfm.bin,BIN,142422792","test.bin,BIN,1234"'
    )
    testdev.com.stb = 32
    testdev.save_waveform_history('test', 1)
    testdev.com.stb = 0
    assert (testdev.com.get_written() ==
            'EXPort:WAVeform:FASTexport ON')
    assert testdev.com.get_written() == 'CHAN:HIST ON'
    assert testdev.com.get_written() == 'SYST:DISP:UPD OFF'
    assert (testdev.com.get_written() ==
            'EXPort:WAVeform:MULTichannel OFF')
    assert (testdev.com.get_written() ==
            'EXPort:WAVeform:SOURce C1W1')
    assert (testdev.com.get_written() ==
            "EXPort:WAVeform:NAME 'C:\\Data\\DavidGraber\\02_waveforms\\test.bin'")
    assert testdev.com.get_written() == 'EXPort:WAVeform:DLOGging ON'
    assert testdev.com.get_written() == '*CLS'
    assert testdev.com.get_written() == 'CHANnel:HISTory:PLAY'
    assert testdev.com.get_written() == '*OPC'
    assert (testdev.com.get_written() ==
            'EXPort:WAVeform:FASTexport OFF')
    assert testdev.com.get_written() == 'SYST:DISP:UPD ON'

    # with timeout
    with pytest.raises(RTO1024Error):
        testdev.save_waveform_history('test', 1)

    # with "file not found"
    testdev.com.put_name(
        "MMEMory:CATalog? '{}'".format(testdev.config.waveforms_path),
        '142422792,147771314176,".,DIR,0","..,DIR,0","DavidGraber,DIR,0","Myriam,DIR,'
        '0"'
    )
    testdev.com.stb = 32
    with pytest.raises(RTO1024Error):
        testdev.save_waveform_history('test', 1)
    testdev.com.stb = 0


def test_run_continuous_acquisition(testdev: RTO1024):
    testdev.run_continuous_acquisition()
    assert testdev.com.get_written() == 'RUN'


def test_run_single_acquisition(testdev: RTO1024):
    testdev.run_single_acquisition()
    assert testdev.com.get_written() == 'SINGle'


def test_stop_acquisition(testdev: RTO1024):
    testdev.stop_acquisition()
    assert testdev.com.get_written() == 'STOP'


def test_prepare_ultra_segmentation(testdev: RTO1024):
    testdev.prepare_ultra_segmentation()


def test_save_configuration(testdev: RTO1024):
    testdev.save_configuration('test_configuration')
    assert testdev.com.get_written() == '*SAV 49'
    assert (
        testdev.com.get_written() ==
        "MMEMory:STOR:STAT 49,'C:\\Data\\DavidGraber\\01_settings\\test_configuration"
        ".dfl'"
    )


def test_load_configuration(testdev: RTO1024):
    testdev.load_configuration('test_configuration')
    assert (
        testdev.com.get_written() ==
        "MMEMory:LOAD:STAT 41,'C:\\Data\\DavidGraber\\01_settings\\test_configuration"
        ".dfl'"
    )
    assert testdev.com.get_written() == '*RCL 41'


def test_get_timestamps(testdev: RTO1024):
    testdev.com.put_name('ACQuire:AVAilable?', '3')
    testdev.com.put_name('CHANnel:WAVeform:HISTory:TSRelative?', '-1.24')
    testdev.com.put_name('CHANnel:WAVeform:HISTory:TSRelative?', '-0.87')
    testdev.com.put_name('CHANnel:WAVeform:HISTory:TSRelative?', '0')
    ts = testdev.get_timestamps()
    assert ts == [-1.24, -0.87, 0]

    # with read errors
    testdev.com.put_name('ACQuire:AVAilable?', '3')
    testdev.com.put_name('CHANnel:WAVeform:HISTory:TSRelative?', '-1.24')
    testdev.com.put_name('CHANnel:WAVeform:HISTory:TSRelative?', '-1.24')
    testdev.com.put_name('CHANnel:WAVeform:HISTory:TSRelative?', '0')
    with pytest.raises(RTO1024Error):
        testdev.get_timestamps()



"""
Tests for the OPC UA communication protocol.
"""

from time import sleep

import pytest

from hvl_ccb.comm import OpcUaCommunication, OpcUaSubHandler
from tests.opctools import DemoServer


@pytest.fixture(scope='module')
def testconfig():
    return {
        'host': 'localhost',
        'port': 14125,
        'endpoint_name': '',
        'sub_handler': MySubHandler(),
        'update_period': 100
    }


@pytest.fixture(scope='module')
def demo_opcua_server():
    opcua_server = DemoServer(100, 'x', 14125)
    opcua_server.start()

    yield opcua_server

    opcua_server.stop()


@pytest.fixture(scope='module')
def connected_comm_protocol(testconfig, demo_opcua_server):
    opc_comm = OpcUaCommunication(testconfig)
    opc_comm.open()
    yield opc_comm
    opc_comm.close()


class MySubHandler(OpcUaSubHandler):

    def __init__(self):
        self.change_counter = 0

    def datachange_notification(self, node, val, data):
        super().datachange_notification(node, val, data)
        print(node, val, data)
        self.change_counter += 1


def test_opcua(testconfig, demo_opcua_server):
    opc_comm = OpcUaCommunication(testconfig)
    assert opc_comm is not None

    opc_comm.open()

    opc_comm.close()


def test_read(connected_comm_protocol, demo_opcua_server):
    demo_opcua_server.add_var('testvar_read', 1.23, True)
    assert connected_comm_protocol.read('testvar_read', 100) == 1.23


def test_write(connected_comm_protocol, demo_opcua_server):
    demo_opcua_server.add_var('testvar_write', 1.23, True)
    connected_comm_protocol.write('testvar_write', 100, 2.04)
    assert connected_comm_protocol.read('testvar_write', 100) == 2.04


def test_init_monitored_nodes(connected_comm_protocol, demo_opcua_server):
    demo_opcua_server.add_var('mon1', 0, False)
    demo_opcua_server.add_var('mon2', 0, False)
    demo_opcua_server.add_var('mon3', 0, False)

    connected_comm_protocol.init_monitored_nodes('mon1', 100)
    connected_comm_protocol.init_monitored_nodes(['mon2', 'mon3'], 100)
    sleep(0.15)


def test_datachange(connected_comm_protocol, demo_opcua_server):
    demo_opcua_server.add_var('test_datachange', 0.1, True)
    connected_comm_protocol.init_monitored_nodes('test_datachange', 100)
    sleep(0.5)

    counter_before = connected_comm_protocol._sub_handler.change_counter
    demo_opcua_server.set_var('test_datachange', 0.2)
    assert demo_opcua_server.get_var('test_datachange') == 0.2
    sleep(0.15)
    counter_after = connected_comm_protocol._sub_handler.change_counter
    assert counter_after == counter_before + 1



"""
Tests for the Elektro Automatik PSI 9000 power supply device classes.
"""

import pytest

from hvl_ccb.dev import (
    PSI9000,
    PSI9000Error,
    PSI9000VisaCommunicationConfig,
    PSI9000Config
)
from tests.masked_comm import MaskedVisaCommunication


@pytest.fixture(scope='module')
def com_config():
    return {
        'interface_type': PSI9000VisaCommunicationConfig.InterfaceType.TCPIP_SOCKET,
        'host': '127.0.0.1',
        'open_timeout': 10,
        'timeout': 50,
    }


@pytest.fixture(scope='module')
def dev_config():
    return {
        'spoll_interval': 0.01,
        'spoll_start_delay': 0,
    }


@pytest.fixture
def testdev(com_config, dev_config):
    com = MaskedVisaCommunication(com_config)
    dev = PSI9000(com, dev_config)
    return dev


def test_psi9000_config():
    # currently there are no non-default config values
    PSI9000Config()

    with pytest.raises(ValueError):
        PSI9000Config(power_limit=-1.0)

    with pytest.raises(ValueError):
        PSI9000Config(voltage_lower_limit=20, voltage_upper_limit=19.0)

    with pytest.raises(ValueError):
        PSI9000Config(voltage_lower_limit=4,
                      voltage_upper_limit=PSI9000Config._VOLTAGE_UPPER_LIMIT + 1)

    with pytest.raises(ValueError):
        PSI9000Config(current_lower_limit=-1.0)

    with pytest.raises(ValueError):
        PSI9000Config(current_lower_limit=4,
                      current_upper_limit=PSI9000Config._CURRENT_UPPER_LIMIT + 1)


def test_instantiation(com_config):
    power_supply = PSI9000(com_config)
    assert power_supply is not None


def test_start_stop(testdev: PSI9000):
    power_supply = testdev
    test_com = power_supply.com

    power_supply.start()

    # stop successful
    test_com.put_name('SYSTem:LOCK:OWNer?', 'NONE')
    test_com.put_name('OUTput?', 'OFF')
    power_supply.stop()

    # stop successful: output and lock on
    test_com.put_name('SYSTem:LOCK:OWNer?', 'REMOTE')
    test_com.put_name('OUTPut?', 'ON')
    test_com.put_name('SOURce:VOLTage?', '0 V')
    test_com.put_name('SOURce:CURRent?', '0 A')
    test_com.put_name('OUTPut?', 'OFF')
    test_com.put_name('SYSTem:LOCK:OWNer?', 'NONE')
    power_supply.stop()


def test_locking(testdev: PSI9000):
    power_supply = testdev
    test_com = power_supply.com

    # lock failed
    test_com.put_name('SYSTem:LOCK:OWNer?', 'NONE')
    test_com.put_name('SYSTem:LOCK:OWNer?', 'NONE')
    with pytest.raises(PSI9000Error):
        power_supply.set_system_lock(True)
    assert test_com.get_written() == 'SYSTem:LOCK ON'

    # lock successful
    test_com.put_name('SYSTem:LOCK:OWNer?', 'NONE')
    test_com.put_name('SYSTem:LOCK:OWNer?', 'REMOTE')
    power_supply.set_system_lock(True)
    assert test_com.get_written() == 'SYSTem:LOCK ON'

    # unlock failed due to still on current
    test_com.put_name('SYSTem:LOCK:OWNer?', 'REMOTE')
    test_com.put_name('MEASure:VOLTage?', '20 V')
    test_com.put_name('MEASure:CURRent?', '100 A')
    with pytest.raises(PSI9000Error):
        power_supply.set_system_lock(False)
    assert test_com.get_written() is None

    # unlock failed
    test_com.put_name('SYSTem:LOCK:OWNer?', 'REMOTE')
    test_com.put_name('MEASure:VOLTage?', '0 V')
    test_com.put_name('MEASure:CURRent?', '0 A')
    test_com.put_name('SYSTem:LOCK:OWNer?', 'REMOTE')
    with pytest.raises(PSI9000Error):
        power_supply.set_system_lock(False)
    assert test_com.get_written() == 'SYSTem:LOCK OFF'

    # unlock successful
    test_com.put_name('SYSTem:LOCK:OWNer?', 'REMOTE')
    test_com.put_name('MEASure:VOLTage?', '0 V')
    test_com.put_name('MEASure:CURRent?', '0 A')
    test_com.put_name('SYSTem:LOCK:OWNer?', 'NONE')
    power_supply.set_system_lock(False)
    assert test_com.get_written() == 'SYSTem:LOCK OFF'

    # illegal answer
    test_com.put_name('SYSTem:LOCK:OWNer?', 'bla')
    with pytest.raises(PSI9000Error):
        power_supply.get_system_lock()


def test_output(testdev: PSI9000):
    power_supply = testdev
    test_com = power_supply.com

    test_com.put_name('OUTPut?', 'ON')
    assert power_supply.get_output()

    test_com.put_name('OUTPut?', 'OFF')
    assert not power_supply.get_output()

    test_com.put_name('OUTPut?', 'bla')
    with pytest.raises(PSI9000Error):
        power_supply.get_output()

    # setting output
    test_com.put_name('OUTPut?', 'ON')
    power_supply.set_output(True)
    assert test_com.get_written() == 'OUTPut ON'

    test_com.put_name('OUTPut?', 'OFF')
    power_supply.set_output(False)
    assert test_com.get_written() == 'OUTPut OFF'

    test_com.put_name('OUTPut?', 'ON')
    with pytest.raises(PSI9000Error):
        power_supply.set_output(False)
    assert test_com.get_written() == 'OUTPut OFF'

    test_com.put_name('OUTPut?', 'OFF')
    with pytest.raises(PSI9000Error):
        power_supply.set_output(True)
    assert test_com.get_written() == 'OUTPut ON'


def test_get_uip_upper_limits(testdev: PSI9000):
    test_com = testdev.com

    # get upper limits
    test_com.put_name('SOURce:VOLTage:LIMit:HIGH?', '80 V')
    test_com.put_name('SOURce:CURRent:LIMit:HIGH?', '1000 A')
    test_com.put_name('SOURce:POWer:LIMit:HIGH?', '10000 W')
    assert testdev.get_uip_upper_limits() == (80, 1000, 10000)


def test_get_ui_lower_limits(testdev: PSI9000):
    test_com = testdev.com

    # get lower limits
    test_com.put_name('SOURce:VOLTage:LIMit:LOW?', '2 V')
    test_com.put_name('SOURce:CURRent:LIMit:LOW?', '5 A')
    assert testdev.get_ui_lower_limits() == (2, 5)


def test_set_upper_limits(testdev: PSI9000):
    test_com = testdev.com

    # set upper limits, success
    test_com.put_name('SOURce:VOLTage:LIMit:LOW?', '2 V')
    test_com.put_name('SOURce:CURRent:LIMit:LOW?', '5 A')
    test_com.put_name('SOURce:VOLTage:LIMit:HIGH?', '10 V')
    test_com.put_name('SOURce:CURRent:LIMit:HIGH?', '20 A')
    test_com.put_name('SOURce:POWer:LIMit:HIGH?', '10000 W')
    testdev.set_upper_limits(10, 20, 10000)
    assert test_com.get_written() == 'SOURce:VOLTage:LIMit:HIGH 10'
    assert test_com.get_written() == 'SOURce:CURRent:LIMit:HIGH 20'
    assert test_com.get_written() == 'SOURce:POWer:LIMit:HIGH 10000'

    # set upper limits, set fail
    test_com.put_name('SOURce:VOLTage:LIMit:LOW?', '2 V')
    test_com.put_name('SOURce:CURRent:LIMit:LOW?', '5 A')
    test_com.put_name('SOURce:VOLTage:LIMit:HIGH?', '10 V')
    test_com.put_name('SOURce:CURRent:LIMit:HIGH?', '200 A')
    test_com.put_name('SOURce:POWer:LIMit:HIGH?', '10000 W')
    with pytest.raises(PSI9000Error):
        testdev.set_upper_limits(10, 20, 10000)
    assert test_com.get_written() == 'SOURce:VOLTage:LIMit:HIGH 10'
    assert test_com.get_written() == 'SOURce:CURRent:LIMit:HIGH 20'
    assert test_com.get_written() == 'SOURce:POWer:LIMit:HIGH 10000'

    # set upper limits, limit fail
    test_com.put_name('SOURce:VOLTage:LIMit:LOW?', '20 V')
    test_com.put_name('SOURce:CURRent:LIMit:LOW?', '5 A')
    with pytest.raises(ValueError):
        testdev.set_upper_limits(10, 20, 10000)
    assert test_com.get_written() is None


def test_set_lower_limits(testdev: PSI9000):
    test_com = testdev.com

    # set lower limits, success
    test_com.put_name('SOURce:VOLTage:LIMit:HIGH?', '80 V')
    test_com.put_name('SOURce:CURRent:LIMit:HIGH?', '2000 A')
    test_com.put_name('SOURce:POWer:LIMit:HIGH?', '10000 W')
    test_com.put_name('SOURce:VOLTage:LIMit:LOW?', '2 V')
    test_com.put_name('SOURce:CURRent:LIMit:LOW?', '5.2 A')
    testdev.set_lower_limits(2, 5.2)
    assert test_com.get_written() == 'SOURce:VOLTage:LIMit:LOW 2'
    assert test_com.get_written() == 'SOURce:CURRent:LIMit:LOW 5.2'

    # set lower limits, set fail
    test_com.put_name('SOURce:VOLTage:LIMit:HIGH?', '80 V')
    test_com.put_name('SOURce:CURRent:LIMit:HIGH?', '2000 A')
    test_com.put_name('SOURce:POWer:LIMit:HIGH?', '10000 W')
    test_com.put_name('SOURce:VOLTage:LIMit:LOW?', '2 V')
    test_com.put_name('SOURce:CURRent:LIMit:LOW?', '50 A')
    with pytest.raises(PSI9000Error):
        testdev.set_lower_limits(2, 5)
    assert test_com.get_written() == 'SOURce:VOLTage:LIMit:LOW 2'
    assert test_com.get_written() == 'SOURce:CURRent:LIMit:LOW 5'

    # set lower limits, limit fail
    test_com.put_name('SOURce:VOLTage:LIMit:HIGH?', '1 V')
    test_com.put_name('SOURce:CURRent:LIMit:HIGH?', '20 A')
    test_com.put_name('SOURce:POWer:LIMit:HIGH?', '10000 W')
    with pytest.raises(ValueError):
        testdev.set_lower_limits(0.5, 500)
    assert test_com.get_written() is None


def test_set_voltage_current(testdev: PSI9000):
    power_supply = testdev
    test_com = power_supply.com

    test_com.put_name('SOURce:VOLTage?', '20.2 V')
    test_com.put_name('SOURce:CURRent?', '123.4 A')
    power_supply.set_voltage_current(20.2, 123.4)
    assert test_com.get_written() == 'SOURce:VOLTage 20.200000'
    assert test_com.get_written() == 'SOURce:CURRent 123.400000'

    test_com.put_name('SOURce:VOLTage?', '20.2 V')
    test_com.put_name('SOURce:CURRent?', '10 A')
    test_com.put_name('SOURce:VOLTage:LIMit:HIGH?', '20 V')
    test_com.put_name('SOURce:CURRent:LIMit:HIGH?', '2000 A')
    test_com.put_name('SOURce:POWer:LIMit:HIGH?', '10000 W')
    test_com.put_name('SOURce:VOLTage:LIMit:LOW?', '1 V')
    test_com.put_name('SOURce:CURRent:LIMit:LOW?', '5 A')
    with pytest.raises(PSI9000Error):
        power_supply.set_voltage_current(20.2, 123.4)
    assert test_com.get_written() == 'SOURce:VOLTage 20.200000'
    assert test_com.get_written() == 'SOURce:CURRent 123.400000'


def test_check_master_slave_config(testdev: PSI9000):
    power_supply = testdev
    test_com = power_supply.com

    # MS disabled
    test_com.put_name('SYSTem:MS:ENABle?', 'OFF')
    with pytest.raises(PSI9000Error):
        power_supply.check_master_slave_config()
    assert test_com.get_written() is None

    # device is not MASTER
    test_com.put_name('SYSTem:MS:ENABle?', 'ON')
    test_com.put_name('SYSTem:MS:LINK?', 'SLAVE')
    with pytest.raises(PSI9000Error):
        power_supply.check_master_slave_config()
    assert test_com.get_written() is None

    # initialization failed
    test_com.put_name('SYSTem:MS:ENABle?', 'ON')
    test_com.put_name('SYSTem:MS:LINK?', 'MASTER')
    test_com.put_name('SYSTem:MS:CONDition?', 'NOT_INIT')
    with pytest.raises(PSI9000Error):
        power_supply.check_master_slave_config()
    assert test_com.get_written() == 'SYSTem:MS:INITialisation'

    # wrong expected voltage, current nominal values
    test_com.put_name('SYSTem:MS:ENABle?', 'ON')
    test_com.put_name('SYSTem:MS:LINK?', 'MASTER')
    test_com.put_name('SYSTem:MS:CONDition?', 'INIT')
    test_com.put_name('SYSTem:MS:NOMinal:VOLTage?', '{} V'.format(
        PSI9000.MS_NOMINAL_VOLTAGE))
    test_com.put_name('SYSTem:MS:NOMinal:CURRent?', '{} A'.format(
        PSI9000.MS_NOMINAL_CURRENT - 1))
    with pytest.raises(PSI9000Error):
        power_supply.check_master_slave_config()
    assert test_com.get_written() == 'SYSTem:MS:INITialisation'

    # all right
    # wrong expected voltage, current nominal values
    test_com.put_name('SYSTem:MS:ENABle?', 'ON')
    test_com.put_name('SYSTem:MS:LINK?', 'MASTER')
    test_com.put_name('SYSTem:MS:CONDition?', 'INIT')
    test_com.put_name('SYSTem:MS:NOMinal:VOLTage?', '{} V'.format(
        PSI9000.MS_NOMINAL_VOLTAGE))
    test_com.put_name('SYSTem:MS:NOMinal:CURRent?', '{} A'.format(
        PSI9000.MS_NOMINAL_CURRENT))
    power_supply.check_master_slave_config()
    assert test_com.get_written() == 'SYSTem:MS:INITialisation'



"""
Labjack Device for hvl_ccb.
Originally developed and tested for LabJack T7-PRO.

Makes use of the LabJack LJM Library Python wrapper.
This wrapper needs an installation of the LJM Library for Windows, Mac OS X or Linux.
Go to:
https://labjack.com/support/software/installers/ljm
and
https://labjack.com/support/software/examples/ljm/python
"""

import logging
from typing import Union

from .base import SingleCommDevice
from ..comm import LJMCommunication
from ..utils.enum import NameEnum


class LabJackError(Exception):
    """
    Errors of the LabJack device.
    """
    pass


class LabJack(SingleCommDevice):
    """
    LabJack Device.
    This class is tested with a LabJack T7-Pro and should work with T4, T7 and DIGIT
    communicating through the LJM Library. Other or older hardware versions and
    variants of LabJack devices are not supported.
    """

    def __init__(self, com, dev_config=None) -> None:
        """
        Constructor for a LabJack Device.

        :param com: Communication protocol object of type
            LJMCommunication. If a configuration (dict or configdataclass) is given,
            a new communication protocol object will be instantiated.
        :param dev_config: There is no device configuration for LabJack yet.
        """
        super().__init__(com, dev_config)

    @staticmethod
    def default_com_cls():
        return LJMCommunication

    def start(self) -> None:
        """
        Start the Device.
        """

        logging.info('Starting Device')

        self.com.open()

    def stop(self) -> None:
        """
        Stop the Device.
        """

        logging.info('Stopping Device')

        self.com.close()

    def get_serial_number(self) -> int:
        """
        Returns the serial number of the connected LabJack.

        :return: Serial number.
        """

        return int(self.com.read_name("SERIAL_NUMBER"))

    def get_sbus_temp(self, number: int) -> float:
        """
        Read the temperature value from a serial SBUS sensor.

        :param number: port number (0..22)
        :return: temperature in Kelvin
        """

        return float(
            self.com.read_name("SBUS{}_TEMP".format(number))
        )

    def get_sbus_rh(self, number: int) -> float:
        """
        Read the relative humidity value from a serial SBUS sensor.

        :param number: port number (0..22)
        :return: relative humidity in %RH
        """

        return float(self.com.read_name("SBUS{}_RH".format(number)))

    def get_ain(self, channel: int) -> float:
        """
        Read the voltage of an analog input.

        :param channel: is the AIN number (0..254)
        :return: the read voltage
        """

        return float(self.com.read_name("AIN{}".format(channel)))

    def set_ain_range(self, channel: int, ain_range: float) -> None:
        r"""
        Set the range of an analog input port.

        Possible values for ``ain_range`` are:

        *  10   => +- 10   V
        *  1    => +- 1    V
        *  0.1  => +- 0.1  V
        *  0.01 => +- 0.01 V

        :param channel: is the AIN number (0..254)
        :param ain_range: is the range specifier
        """

        if ain_range not in (10, 1, 0.1, 0.01):
            raise LabJackError('Not supported range: {}'.format(ain_range))

        self.com.write_name(
            "AIN{}_RANGE".format(channel), ain_range
        )

    def set_ain_resolution(self, channel: int, resolution: int) -> None:
        """
        Set the resolution index of an analog input port.

        For a T7 Pro values between 0-12 are allowed.
        0 will set the resolution index to default value.

        :param channel: is the AIN number (0..254)
        :param resolution: is the resolution index
        """

        if resolution not in range(13):
            raise LabJackError('Not supported resolution index: {}'.format(resolution))

        self.com.write_name(
            'AIN{}_RESOLUTION_INDEX'.format(channel), resolution
        )

    def set_ain_differential(self, pos_channel: int, differential: bool) -> None:
        """
        Sets an analog input to differential mode or not.
        T7-specific: For base differential channels, positive must be even channel
        from 0-12 and negative must be positive+1. For extended channels 16-127,
        see Mux80 datasheet.

        :param pos_channel: is the AIN number (0..12)
        :param differential: True or False
        :raises LabJackError: if parameters are unsupported
        """

        if pos_channel not in range(13):
            raise LabJackError('Not supported pos_channel: {}'.format(pos_channel))

        if pos_channel % 2 != 0:
            raise LabJackError('AIN pos_channel for positive part of differential pair '
                               'must be even: {}'.format(pos_channel))

        neg_channel = pos_channel + 1

        self.com.write_name(
            "AIN{}_NEGATIVE_CH".format(pos_channel),
            neg_channel if differential else 199
        )

    class ThermocoupleType(NameEnum):
        """
        Thermocouple type; NONE means disable thermocouple mode.
        """
        _init_ = 'ef_index'
        NONE = 0
        E = 20
        J = 21
        K = 22
        R = 23
        T = 24
        S = 25
        C = 30
        PT100 = 40
        PT500 = 41
        PT1000 = 42

    class CjcType(NameEnum):
        """
        CJC slope and offset
        """
        _init_ = 'slope offset'
        internal = 1, 0
        lm34 = 55.56, 255.37

    class TemperatureUnit(NameEnum):
        """
        Temperature unit (to be returned)
        """
        _init_ = 'ef_config_a'
        K = 0
        C = 1
        F = 2

    def set_ain_thermocouple(
            self,
            pos_channel: int,
            thermocouple: Union[None, str, ThermocoupleType],
            cjc_address: int = 60050,
            cjc_type: Union[str, CjcType] = CjcType.internal,
            vrange: float = 0.01,
            resolution: int = 10,
            unit: Union[str, TemperatureUnit] = TemperatureUnit.K) -> None:
        """
        Set the analog input channel to thermocouple mode.

        :param pos_channel: is the analog input channel of the positive part of the
            differential pair
        :param thermocouple: None to disable thermocouple mode, or string specifying
            the thermocouple type
        :param cjc_address: modbus register address to read the CJC temperature
        :param cjc_type: determines cjc slope and offset, 'internal' or 'lm34'
        :param vrange: measurement voltage range (10, 1, 0.1, 0.01)
        :param resolution: resolution index (T7 Pro: 0-12)
        :param unit: is the temperature unit to be returned ('K', 'C' or 'F')
        :raises LabJackError: if parameters are unsupported
        """

        if thermocouple is None:
            thermocouple = self.ThermocoupleType.NONE
        else:
            thermocouple = self.ThermocoupleType(thermocouple)

        unit = self.TemperatureUnit(unit)

        cjc_type = self.CjcType(cjc_type)

        self.set_ain_differential(pos_channel=pos_channel, differential=True)
        self.set_ain_range(pos_channel, vrange)
        self.set_ain_resolution(pos_channel, resolution)
        self.set_ain_range(pos_channel + 1, vrange)
        self.set_ain_resolution(pos_channel+1, resolution)

        # specify thermocouple mode
        self.com.write_name(
            "AIN{}_EF_INDEX".format(pos_channel), thermocouple.ef_index
        )

        # specify the units for AIN#_EF_READ_A and AIN#_EF_READ_C (0 = K, 1 = C, 2 = F)
        self.com.write_name(
            "AIN{}_EF_CONFIG_A".format(pos_channel), unit.ef_config_a
        )

        # specify modbus address for cold junction reading CJC
        self.com.write_name(
            "AIN{}_EF_CONFIG_B".format(pos_channel), cjc_address
        )

        # set slope for the CJC reading, typically 1
        self.com.write_name(
            f'AIN{pos_channel}_EF_CONFIG_D', cjc_type.slope
        )

        # set the offset for the CJC reading, typically 0
        self.com.write_name(
            f'AIN{pos_channel}_EF_CONFIG_E', cjc_type.offset
        )

    def read_thermocouple(self, pos_channel: int) -> float:
        """
        Read the temperature of a connected thermocouple.

        :param pos_channel: is the AIN number of the positive pin
        :return: temperature in specified unit
        """

        return round(
                self.com.read_name(
                    "AIN{}_EF_READ_A".format(pos_channel)
                ), 2)



"""Communication protocols subpackage."""

from .base import CommunicationProtocol  # noqa: F401
from .labjack_ljm import (  # noqa: F401
    LJMCommunication,
    LJMCommunicationConfig,
    LJMCommunicationError,
)
from .modbus_tcp import (  # noqa: F401
    ModbusTcpCommunication,
    ModbusTcpConnectionFailedException,
    ModbusTcpCommunicationConfig,
)
from .opc import (  # noqa: F401
    OpcUaCommunication,
    OpcUaCommunicationConfig,
    OpcUaSubHandler,
)
from .serial import SerialCommunication, SerialCommunicationConfig  # noqa: F401
from .visa import (  # noqa: F401
    VisaCommunication,
    VisaCommunicationError,
    VisaCommunicationConfig,
)



"""
Tests for the dev.labjack module.
"""

import pytest

from hvl_ccb.dev import LabJack, LabJackError
from tests.masked_comm import MaskedLJMCommunication


@pytest.fixture(scope="module")
def com_config():
    return {
        'device_type': 'ANY',
        'connection_type': 'ANY',
        'identifier': '-2'  # identifier = -2 specifies LJM DEMO mode, see
        # https://labjack.com/support/software/api/ljm/demo-mode
    }


@pytest.fixture
def started_dev_comm(com_config):
    com = MaskedLJMCommunication(com_config)
    with LabJack(com) as lj:
        yield lj, com


def test_instantiation(com_config):
    lj = LabJack(com_config)
    assert lj is not None


def test_get_serial_number(started_dev_comm):
    lj, com = started_dev_comm
    com.put_name("SERIAL_NUMBER", 1234)
    assert lj.get_serial_number() == 1234


def test_get_sbus_temp(started_dev_comm):
    lj, com = started_dev_comm
    com.put_name("SBUS0_TEMP", 298.15)
    assert lj.get_sbus_temp(0) == 298.15


def test_get_sbus_rh(started_dev_comm):
    lj, com = started_dev_comm
    com.put_name("SBUS0_RH", 40.5)
    assert lj.get_sbus_rh(0) == 40.5


def test_get_ain(started_dev_comm):
    lj, com = started_dev_comm
    com.put_name("AIN0", 1.23)
    assert lj.get_ain(0) == 1.23


def test_set_ain_range(started_dev_comm):
    lj, com = started_dev_comm

    lj.set_ain_range(0, 10)
    assert com.get_written() == ('AIN0_RANGE', 10)

    lj.set_ain_range(0, 0.1)
    assert com.get_written() == ('AIN0_RANGE', 0.1)

    with pytest.raises(LabJackError):
        lj.set_ain_range(0, 0.2)
    assert com.get_written() is None


def test_set_ain_resolution(started_dev_comm):
    lj, com = started_dev_comm

    lj.set_ain_resolution(0, 10)
    assert com.get_written() == ('AIN0_RESOLUTION_INDEX', 10)
    lj.set_ain_resolution(0, 0)
    assert com.get_written() == ('AIN0_RESOLUTION_INDEX', 0)

    with pytest.raises(LabJackError):
        lj.set_ain_resolution(0, -14)
    assert com.get_written() is None

    with pytest.raises(LabJackError):
        lj.set_ain_resolution(0, 13)
    assert com.get_written() is None


def test_set_ain_differential(started_dev_comm):
    lj, com = started_dev_comm

    lj.set_ain_differential(4, True)
    assert com.get_written() == ('AIN4_NEGATIVE_CH', 5)

    lj.set_ain_differential(4, False)
    assert com.get_written() == ('AIN4_NEGATIVE_CH', 199)

    with pytest.raises(LabJackError):
        lj.set_ain_differential(5, True)
    assert com.get_written() is None

    with pytest.raises(LabJackError):
        lj.set_ain_differential(14, True)
    assert com.get_written() is None


def test_set_ain_thermocouple(started_dev_comm):
    lj, com = started_dev_comm

    lj.set_ain_thermocouple(0, None)
    lj.set_ain_thermocouple(0, LabJack.ThermocoupleType.NONE)
    lj.set_ain_thermocouple(0, 'K')
    lj.set_ain_thermocouple(0, LabJack.ThermocoupleType.K)
    with pytest.raises(ValueError):
        lj.set_ain_thermocouple(0, 'B')

    lj.set_ain_thermocouple(0, thermocouple='T', unit='F')
    lj.set_ain_thermocouple(0, thermocouple='T', unit=LabJack.TemperatureUnit.F)
    with pytest.raises(ValueError):
        lj.set_ain_thermocouple(0, 'K', unit='B')

    lj.set_ain_thermocouple(0, thermocouple='T', cjc_type='lm34')
    lj.set_ain_thermocouple(0, thermocouple='T', cjc_type=LabJack.CjcType.lm34)
    with pytest.raises(ValueError):
        lj.set_ain_thermocouple(0, 'K', cjc_type='LM35')


def test_read_thermocouple(started_dev_comm):
    lj, com = started_dev_comm
    com.put_name('AIN0_EF_READ_A', 244.3)
    assert lj.read_thermocouple(0) == 244.30



"""
Supercube Typ B module.
"""

from hvl_ccb.configuration import configdataclass
from . import constants
from .base import (
    SupercubeBase,
    SupercubeOpcUaCommunication,
    SupercubeOpcUaCommunicationConfig
)


@configdataclass
class SupercubeBOpcUaConfiguration(SupercubeOpcUaCommunicationConfig):
    endpoint_name: str = constants.SupercubeOpcEndpoint.B.value


class SupercubeBOpcUaCommunication(SupercubeOpcUaCommunication):
    @staticmethod
    def config_cls():
        return SupercubeBOpcUaConfiguration


class SupercubeB(SupercubeBase):
    """
    Variant B of the Supercube without frequency converter but external safety switches.
    """

    @staticmethod
    def default_com_cls():
        return SupercubeBOpcUaCommunication



"""
Communication protocol for modbus TCP ports. Makes use of the
`pymodbus <https://pymodbus.readthedocs.io/en/latest/>`_ library.
"""

import logging
from typing import List, Union

from IPy import IP
from pymodbus.client.sync import ModbusTcpClient
from pymodbus.exceptions import ConnectionException

from .base import CommunicationProtocol
from ..configuration import configdataclass


class ModbusTcpConnectionFailedException(ConnectionException):
    """
    Exception raised when the connection failed.
    """
    pass


@configdataclass
class ModbusTcpCommunicationConfig:
    """
    Configuration dataclass for :class:`ModbusTcpCommunication`.
    """

    #: Host is the IP address of the connected device.
    host: str

    #: Unit number to be used when connecting with Modbus/TCP. Typically this is used
    #: when connecting to a relay having Modbus/RTU-connected devices.
    unit: int

    #: TCP port
    port: int = 502

    def clean_values(self):
        # host, raises ValueError on its own if not suitable
        IP(self.host)


class ModbusTcpCommunication(CommunicationProtocol):
    """
    Implements the Communication Protocol for modbus TCP.
    """

    def __init__(self, configuration):
        """Constructor for modbus"""
        super().__init__(configuration)

        # create the modbus port specified in the configuration
        logging.debug(
            'Create ModbusTcpClient with host: "{}", Port: "{}", Unit: "{}"'.format(
                self.config.host,
                self.config.port,
                self.config.unit,
            )
        )
        self.client = ModbusTcpClient(
            self.config.host, port=self.config.port, unit=self.config.unit
        )

    @staticmethod
    def config_cls():
        return ModbusTcpCommunicationConfig

    def open(self) -> None:
        """
        Open the Modbus TCP connection.

        :raises ModbusTcpConnectionFailedException: if the connection fails.
        """

        # open the port
        logging.debug('Open Modbus TCP Port.')

        with self.access_lock:
            if not self.client.connect():
                raise ModbusTcpConnectionFailedException

    def close(self):
        """
        Close the Modbus TCP connection.
        """

        # close the port
        logging.debug('Close Modbus TCP Port.')

        with self.access_lock:
            self.client.close()

    def write_registers(self, address: int, values: Union[List[int], int]):
        """
        Write values from the specified address forward.

        :param address: address of the first register
        :param values: list with all values
        """

        logging.debug(
            'Write registers {address} with values {values}'.format(
                address=address,
                values=values,
            )
        )

        with self.access_lock:
            try:
                self.client.write_registers(address=address, values=values,
                                            unit=self.config.unit)
            except ConnectionException as e:
                raise ModbusTcpConnectionFailedException from e

    def read_holding_registers(self, address: int, count: int) -> List[int]:
        """
        Read specified number of register starting with given address and return
        the values from each register.

        :param address: address of the first register
        :param count: count of registers to read
        :return: list of `int` values
        """

        logging.debug(
            'Read holding registers {address} with count {count}.'.format(
                address=address,
                count=count,
            )
        )

        with self.access_lock:
            try:
                registers = self.client.read_holding_registers(
                    address=address,
                    count=count
                ).registers
            except ConnectionException as e:
                raise ModbusTcpConnectionFailedException from e

        logging.debug(
            'Returned holding registers {address}: {registers}'.format(
                address=address,
                registers=registers,
            )
        )

        return registers

    def read_input_registers(self, address: int, count: int) -> List[int]:
        """
        Read specified number of register starting with given address and return
        the values from each register in a list.

        :param address: address of the first register
        :param count: count of registers to read
        :return: list of `int` values
        """

        logging.debug(
            'Read input registers {address} with count {count}.'.format(
                address=address,
                count=count,
            )
        )

        with self.access_lock:
            try:
                registers = self.client.read_input_registers(
                    address=address,
                    count=count
                ).registers
            except ConnectionException as e:
                raise ModbusTcpConnectionFailedException from e

        logging.debug(
            'Returned input registers {address}: {registers}'.format(
                address=address,
                registers=registers,
            )
        )

        return registers



"""
Tests for the base class CommunicationProtocol.
"""

from hvl_ccb.comm import CommunicationProtocol
from hvl_ccb.configuration import configdataclass


@configdataclass
class EmptyConfig:
    pass


class DummyCommmunicationProtocol(CommunicationProtocol):
    @staticmethod
    def config_cls():
        return EmptyConfig

    def open(self):
        pass

    def close(self):
        pass


def test_communication_protocol():
    with DummyCommmunicationProtocol({}) as c:
        assert c is not None



"""
Module with base classes for devices.
"""

from abc import ABC, abstractmethod
from typing import Dict, Callable, Type

from ..comm import CommunicationProtocol
from ..configuration import ConfigurationMixin, configdataclass


class DeviceExistingException(Exception):
    """
    Exception to indicate that a device with that name already exists.
    """

    pass


@configdataclass
class EmptyConfig:
    """
    Empty configuration dataclass that is the default configuration for a Device.
    """
    pass


class Device(ConfigurationMixin, ABC):
    """
    Base class for devices. Implement this class for a concrete device,
    such as measurement equipment or voltage sources.

    Specifies the methods to implement for a device.
    """

    def __init__(self, dev_config=None):
        """
        Constructor for Device.
        """

        super().__init__(dev_config)

    @abstractmethod
    def start(self) -> None:
        """
        Start or restart this Device. To be implemented in the subclass.
        """

        pass  # pragma: no cover

    @abstractmethod
    def stop(self) -> None:
        """
        Stop this Device. To be implemented in the subclass.
        """

        pass  # pragma: no cover

    def __enter__(self):
        self.start()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.stop()

    @staticmethod
    def config_cls():
        return EmptyConfig


class DeviceSequenceMixin(ABC):
    """
    Mixin that can be used on a device or other classes to provide facilities for
    handling multiple devices in a sequence.
    """

    def __init__(self, devices: Dict[str, Device]):
        """
        Constructor for the DeviceSequenceMixin.

        :param devices: is a dictionary of devices to be added to this sequence.
        """

        self._devices = devices  # type: Dict[str, Device]

    def apply_to_devices(self, func: Callable[[Device], object]) -> \
            Dict[str, object]:
        """
        Apply a function to all devices in this sequence.

        :param func: is a function that takes a device as an argument.
        :return: a sequence of objects returned by the called function.
        """

        returned_objects = {}

        for name in self._devices.keys():
            returned_objects[name] = func(self.get_device(name))

        return returned_objects

    def get_device(self, name: str) -> Device:
        """
        Get a device by name.

        :param name: is the name of the device.
        :return: the device object from this sequence.
        """

        return self._devices.get(name)

    def add_device(self, name: str, device: Device) -> None:
        """
        Add a new device to the device sequence.

        :param name: is the name of the device.
        :param device: is the instantiated Device object.
        :raise DeviceExistingException:
        """

        if name in self._devices.keys():
            raise DeviceExistingException

        self._devices[name] = device

    def remove_device(self, name: str) -> Device:
        """
        Remove a device from this sequence and return the object.

        :param name: is the name of the device.
        :return: device object.
        """

        return self._devices.pop(name, None)

    def start(self) -> None:
        """
        Start all devices in this sequence in their added order.
        """

        for name in self._devices.keys():
            self.get_device(name).start()

    def stop(self) -> None:
        """
        Stop all devices in this sequence in their reverse order.
        """

        for name in reversed(list(self._devices.keys())):
            self.get_device(name).stop()


class SingleCommDevice(Device, ABC):
    """
    Base class for devices with a single communication protocol.
    """

    # Omitting typing hint `com: CommunicationProtocol` on purpose
    # to enable PyCharm autocompletion for subtypes.
    def __init__(self, com, dev_config=None) -> None:
        """
        Constructor for Device. Links the communication protocol and provides a
        configuration for the device.

        :param com: Communication protocol to be used with
            this device. Can be of type: - CommunicationProtocol instance, - dictionary
            with keys and values to be used as configuration together with the
            default communication protocol, or - @configdataclass to be used together
            with the default communication protocol.

        :param dev_config: configuration of the device. Can be:
            - None: empty configuration is used, or the specified config_cls()
            - @configdataclass decorated class
            - Dictionary, which is then used to instantiate the specified config_cls()
        """

        super().__init__(dev_config)

        if isinstance(com, CommunicationProtocol):
            self._com = com
        else:
            self._com = self.default_com_cls()(com)

    @staticmethod
    @abstractmethod
    def default_com_cls() -> Type[CommunicationProtocol]:
        """
        Get the class for the default communication protocol used with this device.

        :return: the type of the standard communication protocol for this device
        """
        pass  # pragma: no cover

    @property
    def com(self):
        """
        Get the communication protocol of this device.

        :return: an instance of CommunicationProtocol subtype
        """
        return self._com

    def start(self) -> None:
        """
        Open the associated communication protocol.
        """

        self.com.open()

    def stop(self) -> None:
        """
        Close the associated communication protocol.
        """

        self.com.close()



"""
Tests for configuration.py: Mixin and class decorator configdataclass
"""

import os
from typing import Union

import pytest

from hvl_ccb.configuration import ConfigurationMixin, configdataclass


@configdataclass
class MyConfiguration:
    field1: int
    field2: str = 'hello'

    def clean_values(self):
        if self.field1 == 0:
            raise ValueError('field1 is not allowed to be 0')

        if self.field2 == '':
            self.force_value('field2', 'EMPTY')


@configdataclass
class SimpleConfiguration:
    field1: str = 'field1'


class MyClassHasConfiguration(ConfigurationMixin):

    @staticmethod
    def config_cls():
        return MyConfiguration


# TESTS
def test_inheritance():
    @configdataclass
    class SubMyConfiguration(MyConfiguration):
        field1: int = 3
        field3: int = 5

    assert SubMyConfiguration.optional_defaults() == {
        'field1': 3,
        'field2': 'hello',
        'field3': 5,
    }

    SubMyConfiguration()

    with pytest.raises(AttributeError):
        @configdataclass
        class SubFailingConfiguration(MyConfiguration):
            def keys(self):
                pass


def test_configdataclass():
    SimpleConfiguration()

    # test own clean_values implementation
    with pytest.raises(ValueError):
        MyConfiguration(0)

    # test force_value
    config = MyConfiguration(1, field2='')
    assert config.field2 == 'EMPTY'

    # test convenience functions
    assert MyConfiguration.optional_defaults() == {'field2': 'hello'}
    assert MyConfiguration.required_keys() == ['field1']
    assert MyConfiguration.keys() == ['field1', 'field2']

    # typing is enforced
    with pytest.raises(TypeError):
        MyConfiguration('test')


def test_configdataclass_with_generic():
    @configdataclass
    class ConfigurationWithGeneric:
        number: Union[int, float]

    with pytest.raises(TypeError):
        ConfigurationWithGeneric("1")
    assert ConfigurationWithGeneric(1).number == 1
    assert ConfigurationWithGeneric(1.0).number == 1.0


def test_configdataclass_fromdict():
    test_config = {
        'field1': 1,
        'field2': 'otherstring'
    }

    my_class = MyClassHasConfiguration(test_config)
    assert my_class.config.field1 == 1
    assert my_class.config.field2 == 'otherstring'

    test_config_2 = {
        'field3': 3,
        'field4': 4
    }

    # superfluous fields are not allowed
    with pytest.raises(TypeError):
        MyClassHasConfiguration(test_config_2)

    # non-configdataclasses are not allowed
    with pytest.raises(TypeError):
        class WrongDefaultConfigDataclass(ConfigurationMixin):
            @staticmethod
            def config_cls():
                class NotAConfigDataclass:
                    field1: int = 0
                return NotAConfigDataclass

        WrongDefaultConfigDataclass({'field1': 1})


def test_configuration_mixin():
    test_config = MyConfiguration(1)
    my_class = MyClassHasConfiguration(test_config)
    assert my_class.config == test_config

    with pytest.raises(TypeError):
        MyClassHasConfiguration(object)


def test_json_save_load():
    """
    Test the JSON configuration save and load feature.
    """

    c = MyClassHasConfiguration({'field1': 1})
    c.configuration_save_json('test.json')

    # create new protocol from that JSON configuration
    d = MyClassHasConfiguration.from_json('test.json')

    # clean up, delete file
    os.remove('test.json')

    assert c.config == d.config


def test_unfrozen():
    @configdataclass(frozen=False)
    class Test:
        field1: int = 3

    test = Test()
    assert test.field1 == 3
    test.field1 = 5
    assert test.field1 == 5

    # attention: typing is not enforced!
    test.field1 = 'bla'
    assert test.field1 == 'bla'



"""Devices subpackage."""

from .base import (  # noqa: F401
    Device,
    SingleCommDevice,
    DeviceSequenceMixin,
    DeviceExistingException,
)
from .ea_psi9000 import (  # noqa: F401
    PSI9000,
    PSI9000Config,
    PSI9000VisaCommunication,
    PSI9000VisaCommunicationConfig,
    PSI9000Error,
)
from .labjack import (  # noqa: F401
    LabJack,
    LabJackError,
)
from .mbw973 import (  # noqa: F401
    MBW973,
    MBW973Config,
    MBW973ControlRunningException,
    MBW973PumpRunningException,
    MBW973Error,
    MBW973SerialCommunication,
    MBW973SerialCommunicationConfig,
)
from .rs_rto1024 import (  # noqa: F401
    RTO1024,
    RTO1024Error,
    RTO1024Config,
    RTO1024VisaCommunication,
    RTO1024VisaCommunicationConfig,
)
from .se_ils2t import (  # noqa: F401
    ILS2T,
    ILS2TConfig,
    ILS2TException,
    ILS2TModbusTcpCommunication,
    ILS2TModbusTcpCommunicationConfig,
    IoScanningModeValueError,
    ScalingFactorValueError,
)
from .visa import (  # noqa: F401
    VisaDevice,
    VisaDeviceConfig,
)



"""
Communication protocol for LabJack using the LJM Library.
Originally developed and tested for LabJack T7-PRO.

Makes use of the LabJack LJM Library Python wrapper.
This wrapper needs an installation of the LJM Library for Windows, Mac OS X or Linux.
Go to:
https://labjack.com/support/software/installers/ljm
and
https://labjack.com/support/software/examples/ljm/python
"""

import logging
from typing import Union, Sequence, Tuple

from labjack import ljm

from .base import CommunicationProtocol
from ..configuration import configdataclass
from ..utils.enum import AutoNumberNameEnum


class LJMCommunicationError(Exception):
    """
    Errors coming from LJMCommunication.
    """
    pass


@configdataclass
class LJMCommunicationConfig:
    """
    Configuration dataclass for :class:`LJMCommunication`.
    """

    class DeviceType(AutoNumberNameEnum):
        """
        LabJack device type.
        """
        ANY = ()
        T7 = ()
        T4 = ()
        DIGIT = ()

    #: Can bei either string 'ANY', 'T7', 'T4', 'DIGIT' or of enum :class:`DeviceType`.
    device_type: (str, DeviceType) = 'ANY'

    class ConnectionType(AutoNumberNameEnum):
        """
        LabJack connection type.
        """
        ANY = ()
        USB = ()
        TCP = ()
        ETHERNET = ()
        WIFI = ()

    #: Can be either string or of enum :class:`ConnectionType`.
    connection_type: (str, ConnectionType) = 'ANY'

    identifier: str = 'ANY'
    """
    The identifier specifies information for the connection to be used. This can
    be an IP address, serial number, or device name. See the LabJack docs (
    https://labjack.com/support/software/api/ljm/function-reference/ljmopens/\
identifier-parameter) for more information.
    """

    def clean_values(self) -> None:
        """
        Performs value checks on device_type and connection_type.
        """
        if not isinstance(self.device_type, self.DeviceType):
            self.DeviceType(self.device_type)

        if not isinstance(self.connection_type, self.ConnectionType):
            self.ConnectionType(self.connection_type)


class LJMCommunication(CommunicationProtocol):
    """
    Communication protocol implementing the LabJack LJM Library Python wrapper.
    """

    def __init__(self, configuration) -> None:
        """
        Constructor for LJMCommunication.
        """
        super().__init__(configuration)

        # reference to the ctypes handle
        self._handle = None

        self.logger = logging.getLogger(__name__)

    @staticmethod
    def config_cls():
        return LJMCommunicationConfig

    def open(self) -> None:
        """
        Open the communication port.
        """

        self.logger.info('Open connection')

        # open connection and store handle
        # may throw 1227 LJME_DEVICE_NOT_FOUND if device is not found
        try:
            with self.access_lock:
                self._handle = ljm.openS(
                    str(self.config.device_type),
                    str(self.config.connection_type),
                    str(self.config.identifier),
                )
        except ljm.LJMError as e:
            self.logger.error(e)
            # only catch "1229 LJME_DEVICE_ALREADY_OPEN", never observed
            if e.errorCode == 1229:
                return
            raise LJMCommunicationError from e

    def close(self) -> None:
        """
        Close the communication port.
        """

        self.logger.info('Closing connection')

        try:
            with self.access_lock:
                ljm.close(self._handle)
        except ljm.LJMError as e:
            self.logger.error(e)
            # only catch "1224 LJME_DEVICE_NOT_OPEN", thrown on invalid handle
            if e.errorCode == 1224:
                return
            raise LJMCommunicationError from e

    def __del__(self) -> None:
        """
        Finalizer, closes port
        """

        self.close()

    def read_name(self, *names: str) -> Union[str, Tuple[str]]:
        """
        Read one or more inputs by name.

        :param names: one or more names to read out from the LabJack
        :return: answer of the LabJack, either single answer or multiple answers in a
            tuple
        """

        # Errors that can be returned here:
        # 1224 LJME_DEVICE_NOT_OPEN if the device is not open
        # 1239 LJME_DEVICE_RECONNECT_FAILED if the device was opened, but connection
        #   lost

        with self.access_lock:
            try:
                if len(names) == 1:
                    return ljm.eReadName(self._handle, names[0])
                else:
                    return tuple(ljm.eReadNames(self._handle, len(names), names))
            except ljm.LJMError as e:
                self.logger.error(e)
                raise LJMCommunicationError from e

    def write_name(self,
                   name: Union[Sequence[str], str],
                   value: Union[Sequence[object], object]) -> None:
        """
        Write one value to a named output.

        :param name: String or with name of LabJack IO
        :param value: is the value to write to the named IO port
        """

        with self.access_lock:
            try:
                ljm.eWriteName(self._handle, name, value)
            except ljm.LJMError as e:
                self.logger.error(e)
                raise LJMCommunicationError from e

    def write_names(self, names: Sequence[str], values: Sequence[object]) -> None:
        """
        Write more than one value at once to named outputs.

        :param names: is a sequence of strings with names of LabJack IO
        :param values: is a sequence of values corresponding to the list of names
        """

        with self.access_lock:
            try:
                ljm.eWriteNames(self._handle, len(names), names, values)
            except ljm.LJMError as e:
                self.logger.error(e)
                raise LJMCommunicationError from e

    def write_address(self,
                      address: Union[Sequence[int], int],
                      value: Union[Sequence[object], object]) -> None:
        """
        **NOT IMPLEMENTED.**
        Write one or more values to Modbus addresses.

        :param address: One or more Modbus address on the LabJack.
        :param value: One or more values to be written to the addresses.
        """

        raise NotImplementedError
        # TODO: Implement function to write on addresses. Problem so far: I also need
        #  to bring in the data types (INT32, FLOAT32...)



"""
Module with base classes for communication protocols.
"""

from abc import ABC, abstractmethod
from threading import Lock

from ..configuration import ConfigurationMixin


class CommunicationProtocol(ConfigurationMixin, ABC):
    """
    Communication protocol abstract base class.

    Specifies the methods to implement for communication protocol, as well as
    implements some default settings and checks.
    """

    def __init__(self, config) -> None:
        """
        Constructor for CommunicationProtocol. Takes a configuration dict or
        configdataclass as single parameter.

        :param config: Configdataclass or dictionary to be used with the default
            config dataclass.
        """

        super().__init__(config)

        #: Access lock to use with context manager when
        #: accessing the communication protocol (thread safety)
        self.access_lock = Lock()

    @abstractmethod
    def open(self):
        """
        Open communication protocol
        """
        pass  # pragma: no cover

    @abstractmethod
    def close(self):
        """
        Close the communication protocol
        """
        pass  # pragma: no cover

    def __enter__(self):
        self.open()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()



"""
Tests for .comm sub-package
"""
import dataclasses
import time

import pytest

from hvl_ccb.comm import SerialCommunication, SerialCommunicationConfig


@pytest.fixture(scope="module")
def testconfig():
    return SerialCommunicationConfig(
        port="loop://?logging=debug",
        baudrate=115200,
        parity=SerialCommunicationConfig.Parity.NONE,
        stopbits=SerialCommunicationConfig.Stopbits.ONE,
        bytesize=SerialCommunicationConfig.Bytesize.EIGHTBITS,
        terminator=b'\r\n',
        timeout=0.2
    )


def test_serial_config(testconfig):
    with pytest.raises(ValueError):
        dataclasses.replace(testconfig, parity='B')

    with pytest.raises(ValueError):
        dataclasses.replace(testconfig, stopbits=2.5)

    with pytest.raises(ValueError):
        dataclasses.replace(testconfig, bytesize=9)


def _decode_terminator(testconfig):
    return testconfig.terminator.decode(SerialCommunication.ENCODING)


def test_timeout(testconfig):
    with SerialCommunication(testconfig) as sc:
        started_at = time.time()
        assert sc.read_text() == ''
        elapsed = time.time() - started_at
        timeout = testconfig.timeout
        assert elapsed >= timeout
        assert elapsed < 1.25 * timeout


def _test_serial_communication(testconfig, sc):
    # send something
    test_strings = ["Test message 1",
                    "testmessage2",
                    "190testmessage: 3",
                    ]

    for t in test_strings:
        # send line
        sc.write_text(t)

        # read back line
        answer = sc.read_text()

        assert answer == t + _decode_terminator(testconfig)


def test_serial_communication(testconfig):
    """
    Tests SerialCommunication
    """

    # manually open/close port
    sc = SerialCommunication(testconfig)
    assert sc is not None
    sc.open()
    _test_serial_communication(testconfig, sc)
    sc.close()

    # or use with statement
    with SerialCommunication(testconfig) as sc:
        assert sc is not None
        _test_serial_communication(testconfig, sc)



"""
Device class for controlling a MBW 973 SF6 Analyzer over a serial connection.

The MBW 973 is a gas analyzer designed for gas insulated switchgear and measures
humidity, SF6 purity and SO2 contamination in one go.
Manufacturer homepage: https://www.mbw.ch/products/sf6-gas-analysis/973-sf6-analyzer/
"""

import logging
from threading import Timer
from typing import Dict, Callable, Type

from serial import SerialException

from .base import SingleCommDevice
from ..comm import SerialCommunication, SerialCommunicationConfig
from ..configuration import configdataclass


class MBW973Error(Exception):
    """
    General error with the MBW973 dew point mirror device.
    """
    pass


class MBW973ControlRunningException(MBW973Error):
    """
    Error indicating there is still a measurement running, and a new one cannot be
    started.
    """
    pass


class MBW973PumpRunningException(MBW973Error):
    """
    Error indicating the pump of the dew point mirror is still recovering gas,
    unable to start a new measurement.
    """
    pass


@configdataclass
class MBW973SerialCommunicationConfig(SerialCommunicationConfig):
    #: Baudrate for MBW973 is 9600 baud
    baudrate: int = 9600

    #: MBW973 does not use parity
    parity: (str, SerialCommunicationConfig.Parity) = \
        SerialCommunicationConfig.Parity.NONE

    #: MBW973 does use one stop bit
    stopbits: (int, SerialCommunicationConfig.Stopbits) = \
        SerialCommunicationConfig.Stopbits.ONE

    #: One byte is eight bits long
    bytesize: (int, SerialCommunicationConfig.Bytesize) = \
        SerialCommunicationConfig.Bytesize.EIGHTBITS

    #: The terminator is only CR
    terminator: bytes = b'\r'

    #: use 3 seconds timeout as default
    timeout: (int, float) = 3


class MBW973SerialCommunication(SerialCommunication):
    """
    Specific communication protocol implementation for the MBW973 dew point mirror.
    Already predefines device-specific protocol parameters in config.
    """

    @staticmethod
    def config_cls():
        return MBW973SerialCommunicationConfig


@configdataclass
class MBW973Config:
    """
    Device configuration dataclass for MBW973.
    """

    #: Polling period for `is_done` status queries [in seconds].
    polling_interval: (int, float) = 2

    def clean_values(self):
        if self.polling_interval <= 0:
            raise ValueError('Polling interval needs to be positive.')


class MBW973(SingleCommDevice):
    """
    MBW 973 dew point mirror device class.
    """

    def __init__(self, com, dev_config=None):

        # Call superclass constructor
        super().__init__(com, dev_config)

        # polling timer thread
        self.polling_timer = Poller(self.config.polling_interval, self.is_done)

        # is done with dew point = True, new measurement sample required and
        # not ready yet = False
        self.is_done_with_measurements = True

        # dict telling what measurement options are selected
        self.measurement_options = {
            'dewpoint': True,
            'SF6_Vol': False,
        }

        self.last_measurement_values = {}

    @staticmethod
    def default_com_cls():
        return MBW973SerialCommunication

    @staticmethod
    def config_cls():
        return MBW973Config

    def start(self) -> None:
        """
        Start this device. Opens the communication protocol and retrieves the
        set measurement options from the device.
        """

        logging.info("Starting device " + str(self))

        # TODO: re-factor this error-handling to SerialCommunication.open()?
        try:
            self.com.open()
        except SerialException as exc:
            # only catch "port already open"
            if str(exc) != 'Port is already open.':
                raise MBW973Error from exc

        # check test options
        self.write('HumidityTest?')
        self.measurement_options['dewpoint'] = bool(self.read_int())

        self.write('SF6PurityTest?')
        self.measurement_options['SF6_Vol'] = bool(self.read_int())

    def stop(self) -> None:
        """
        Stop the device. Closes also the communication protocol.
        """

        logging.info("Stopping device " + str(self))
        self.com.close()

    def write(self, value) -> None:
        """
        Send `value` to `self.com`.

        :param value: Value to send, converted to `str`.
        """
        self.com.write_text(str(value))

    def read(self, cast_type: Type = str):
        """
        Read value from `self.com` and cast to `cast_type`.
        Raises `ValueError` if read text (`str`) is not convertible to `cast_type`,
        e.g. to `float` or to `int`.

        :return: Read value of `cast_type` type.
        """
        return cast_type(self.com.read_text())

    def read_float(self) -> float:
        """
        Convenience wrapper for `self.read()`, with typing hint for return value.

        :return: Read `float` value.
        """
        return self.read(float)

    def read_int(self) -> int:
        """
        Convenience wrapper for `self.read()`, with typing hint for return value.

        :return: Read `int` value.
        """
        return self.read(int)

    def is_done(self) -> bool:
        """
        Poll status of the dew point mirror and return True, if all
        measurements are done.

        :return: True, if all measurements are done; False otherwise.
        """

        # assume everything is done
        done = True

        if self.measurement_options['dewpoint']:
            # ask if done with DP
            self.write('DoneWithDP?')
            done = done and bool(self.read_int())

        if self.measurement_options['SF6_Vol']:
            # ask if done with SF6 volume measurement
            self.write('SF6VolHold?')
            done = done and bool(self.read_int())

        self.is_done_with_measurements = done

        if self.is_done_with_measurements:
            self.polling_timer.stop()
            self.read_measurements()

        return self.is_done_with_measurements

    def start_control(self) -> None:
        """
        Start dew point control to acquire a new value set.
        """

        # send control?
        self.write('control?')

        if self.read_float():
            raise MBW973ControlRunningException

        # send Pump.on? to check, whether gas is still being pumped back
        self.write('Pump.on?')

        if self.read_float():
            raise MBW973PumpRunningException

        # start control of device
        self.write('control=1')
        logging.info('Starting dew point control')
        self.is_done_with_measurements = False
        self.polling_timer.start()

    def read_measurements(self) -> Dict[str, float]:
        """
        Read out measurement values and return them as a dictionary.

        :return: Dictionary with values.
        """

        self.write('Fp?')
        frostpoint = self.read_float()

        self.write('Fp1?')
        frostpoint_ambient = self.read_float()

        self.write('Px?')
        pressure = self.read_float()

        self.write('PPMv?')
        ppmv = self.read_float()

        self.write('PPMw?')
        ppmw = self.read_float()

        self.write('SF6Vol?')
        sf6_vol = self.read_float()

        values = {
            'frostpoint': frostpoint,
            'frostpoint_ambient': frostpoint_ambient,
            'pressure': pressure,
            'ppmv': ppmv,
            'ppmw': ppmw,
            'sf6_vol': sf6_vol,
        }

        logging.info('Read out values')
        logging.info(values)

        self.last_measurement_values = values

        return values

    def set_measuring_options(self, humidity: bool = True,
                              sf6_purity: bool = False) -> None:
        """
        Send measuring options to the dew point mirror.

        :param humidity: Perform humidity test or not?
        :param sf6_purity: Perform SF6 purity test or not?
        """

        self.write('HumidityTest={}'.format(1 if humidity else 0))
        self.write('SF6PurityTest={}'.format(1 if sf6_purity else 0))

        self.measurement_options['dewpoint'] = humidity
        self.measurement_options['SF6_Vol'] = sf6_purity


class Poller:
    """
    Wrapper for the threading.Timer class to periodically poll data.
    """

    def __init__(self, period: float, callback: Callable) -> None:
        """
        Constructor for Poller.

        :param period: Period in seconds.
        :param callback: Callback function triggered every `period` seconds.
        """

        self.period = period
        self.callback = callback

        self.t = Timer(period, self.timer_callback)

        # status: False: stopped, True: runs
        self.status = False

    def timer_callback(self) -> None:
        """
        Callback method that is called every time the timer elapses. It calls
        the specified user callback function and
        restarts the timer.
        """

        # call external callback function
        logging.debug('Timer elapsed, call specified callback function: {}'
                      .format(self.callback))
        self.callback()

        logging.debug('Re-initialize timer')
        self.t = Timer(self.period, self.timer_callback)

        logging.debug('Re-start timer')
        if self.status:
            self.t.start()

    def start(self) -> None:
        """
        Start the polling timer.
        """

        logging.debug('Start polling timer')
        self.status = True
        self.t.start()

    def stop(self) -> None:
        """
        Stop the polling timer.
        """

        logging.debug('Stop polling timer')
        self.status = False
        self.t.cancel()



"""
Communication protocol implementing an OPC UA connection.
This protocol is used to interface with the "Supercube" PLC from Siemens.
"""

import logging
from typing import Iterable, Union

from opcua import Client, Node, Subscription
from opcua.ua import NodeId, DataValue

from .base import CommunicationProtocol
from ..configuration import configdataclass


class OpcUaSubHandler:
    """
    Base class for subscription handling of OPC events and data change events.
    Override methods from this class to add own handling capabilities.

    To receive events from server for a subscription
    data_change and event methods are called directly from receiving thread.
    Do not do expensive, slow or network operation there. Create another
    thread if you need to do such a thing.
    """

    def datachange_notification(self, node, val, data):
        logging.getLogger(__name__).debug(
            'OPCUA Datachange event: {} to value {}'.format(node, val)
        )

    def event_notification(self, event):
        logging.getLogger(__name__).debug('OPCUA Event: {}'.format(event))


@configdataclass
class OpcUaCommunicationConfig:
    """
    Configuration dataclass for OPC UA Communciation.
    """

    #: Hostname or IP-Address of the OPC UA server.
    host: str

    #: Endpoint of the OPC server, this is a path like 'OPCUA/SimulationServer'
    endpoint_name: str

    #: Port of the OPC UA server to connect to.
    port: int = 4840

    #: object to use for handling subscriptions.
    sub_handler: OpcUaSubHandler = OpcUaSubHandler()

    #: Update period for generating datachange events in OPC UA [milli seconds]
    update_period: int = 500


class OpcUaCommunication(CommunicationProtocol):
    """
    Communication protocol implementing an OPC UA connection.
    Makes use of the package python-opcua.
    """

    def __init__(self, config) -> None:
        """
        Constructor for OpcUaCommunication.

        :param config: is the configuration dictionary.
        """

        super().__init__(config)

        self.logger = logging.getLogger(__name__)

        url = (
            'opc.tcp://{host}:{port}/{endpoint}'.format(
                host=self.config.host,
                port=self.config.port,
                endpoint=self.config.endpoint_name,
            )
        )

        self.logger.info('Create OPC UA client to URL: {}'.format(url))

        self._client = Client(url)

        # the objects node exists on evere OPC UA server and is the root for all
        # objects.
        self._objects_node = None  # type: Node

        # subscription handler
        self._sub_handler = self.config.sub_handler

        # subscription object
        self._subscription = None  # type: Subscription

    @staticmethod
    def config_cls():
        return OpcUaCommunicationConfig

    def open(self) -> None:
        """
        Open the communication to the OPC UA server.
        """

        self.logger.info('Open connection to OPC server.')
        with self.access_lock:
            self._client.connect()
            # in example from opcua, load_type_definitions() is called after connect(
            # ). However, this raises ValueError when connecting to Siemens S7,
            # and no problems are detected omitting this call.
            # self._client.load_type_definitions()
            self._objects_node = self._client.get_objects_node()
            self._subscription = self._client.create_subscription(
                self.config.update_period, self._sub_handler)

    def close(self) -> None:
        """
        Close the connection to the OPC UA server.
        """

        self.logger.info('Close connection to OPC server.')
        with self.access_lock:
            self._subscription.delete()
            self._client.disconnect()

    def read(self, node_id, ns_index):
        """
        Read a value from a node with id and namespace index.

        :param node_id: the ID of the node to read the value from
        :param ns_index: the namespace index of the node
        :return: the value of the node object.
        """

        with self.access_lock:
            return self._client.get_node(
                NodeId(identifier=node_id, namespaceidx=ns_index)).get_value()

    def write(self, node_id, ns_index, value) -> None:
        """
        Write a value to a node with name ``name``.

        :param node_id: the id of the node to write the value to.
        :param ns_index: the namespace index of the node.
        :param value: the value to write.
        """

        with self.access_lock:
            self._client.get_node(
                NodeId(identifier=node_id, namespaceidx=ns_index)
            ).set_value(DataValue(value))

    def init_monitored_nodes(self, node_id: Union[str, Iterable[str]],
                             ns_index: int) -> None:
        """
        Initialize monitored nodes.

        :param node_id: one or more strings of node IDs.
        :param ns_index: the namespace index the nodes belong to.
        """

        if isinstance(node_id, str):
            ids = [node_id]
        else:
            ids = node_id

        nodes = []
        for id_ in ids:
            nodes.append(self._client.get_node(
                NodeId(identifier=id_, namespaceidx=ns_index)
            ))

        with self.access_lock:
            self._subscription.subscribe_data_change(nodes)



"""
Masked version of LJM Communication for testing purposes.
"""

from collections import defaultdict
from queue import Queue
from typing import Union, Sequence

from hvl_ccb import comm


class MaskedLJMCommunication(comm.LJMCommunication):
    """
    Masks the LJMCommunication protocol disabling read and write to be able to test
    the functionality of devices without
    """

    def __init__(self, configuration):
        super().__init__(configuration)

        self._read_buffer = defaultdict(Queue)
        self._write_buffer = Queue()

    def read_name(self,  *names: str):

        out = [
            self._read_buffer[name].get() if not self._read_buffer[name].empty() else 0
            for name in names
        ]

        return out[0] if len(out) == 1 else out

    def write_name(self,
                   name: Union[Sequence[str], str],
                   value: Union[Sequence[object], object]):

        if isinstance(name, str):
            name = [name]
            value = [value]

        for pair in zip(name, value):
            self._write_buffer.put(pair)

    def write_address(self,
                      address: Union[Sequence[int], int],
                      value: Union[Sequence[object], object]):
        pass

    def put_name(self, name: str, value: object):
        self._read_buffer[name].put(value)

    def get_written(self):
        return self._write_buffer.get() if not self._write_buffer.empty() else None



"""
Device class for controlling a Schneider Electric ILS2T stepper drive over modbus TCP.
"""

import logging
from datetime import timedelta
from enum import Flag, IntEnum
from time import sleep
from typing import Dict, List, Any

import aenum
from bitstring import BitArray
from pymodbus.constants import Endian
from pymodbus.payload import BinaryPayloadDecoder, BinaryPayloadBuilder

from .base import SingleCommDevice
from ..comm import (
    ModbusTcpCommunication,
    ModbusTcpConnectionFailedException,
    ModbusTcpCommunicationConfig,
)
from ..configuration import configdataclass


class ILS2TException(Exception):
    """
    Exception to indicate problems with the SE ILS2T stepper motor.
    """

    pass


class IoScanningModeValueError(ILS2TException):
    """
    Exception to indicate that the selected IO scanning mode is invalid.
    """

    pass


class ScalingFactorValueError(ILS2TException):
    """
    Exception to indicate that a scaling factor value is invalid.
    """

    pass


@configdataclass
class ILS2TModbusTcpCommunicationConfig(ModbusTcpCommunicationConfig):
    """
    Configuration dataclass for Modbus/TCP communciation specific for the Schneider
    Electric ILS2T stepper motor.
    """

    #: The unit has to be 255 such that IO scanning mode works.
    unit: int = 255


class ILS2TModbusTcpCommunication(ModbusTcpCommunication):
    """
    Specific implementation of Modbus/TCP for the Schneider Electric ILS2T stepper
    motor.
    """

    @staticmethod
    def config_cls():
        return ILS2TModbusTcpCommunicationConfig


@configdataclass
class ILS2TConfig:
    """
    Configuration for the ILS2T stepper motor device.
    """

    #: initial maximum RPM for the motor, can be set up to 3000 RPM. The user is
    #: allowed to set a new max RPM at runtime using :meth:`ILS2T.set_max_rpm`,
    #: but the value must never exceed this configuration setting.
    rpm_max_init: int = 1500


class ILS2T(SingleCommDevice):
    """
    Schneider Electric ILS2T stepper drive class.
    """

    class RegDatatype(aenum.Enum):
        """
        Modbus Register Datatypes

        From the manual of the drive:

        =========== =========== ============== =============
        datatype    byte        min            max
        =========== =========== ============== =============
        INT8        1 Byte      -128           127
        UINT8       1 Byte      0              255
        INT16       2 Byte      -32_768        32_767
        UINT16      2 Byte      0              65_535
        INT32       4 Byte      -2_147_483_648 2_147_483_647
        UINT32      4 Byte      0              4_294_967_295
        BITS        just 32bits N/A            N/A
        =========== =========== ============== =============

        """
        _init_ = 'min max'
        INT32 = -2_147_483_648, 2_147_483_647

        def is_in_range(self, value: int) -> bool:
            return self.min <= value <= self.max

    class RegAddr(IntEnum):
        """
        ILS2T Modbus Register Adresses
        """
        POSITION = 7706  # INT32 position of the motor in user defined units
        IO_SCANNING = 6922  # BITS start register for IO scanning control
        # and status
        TEMP = 7200  # INT16 temperature of motor
        VOLT = 7198  # UINT16 dc voltage of motor
        SCALE = 1550  # INT32 user defined steps per revolution
        ACCESS_ENABLE = 282  # BITS not documented register
        # to enable access via IO scanning
        JOGN_FAST = 10506  # UINT16 revolutions per minute for fast Jog (1 to 3000)
        JOGN_SLOW = 10504  # UINT16 revolutions per minute
        # for slow Jog (1 to 3000)

        RAMP_TYPE = 1574  # INT16 ramp type, 0: linear / -1: motor optimized
        RAMP_ACC = 1556  # UINT32 acceleration
        RAMP_DECEL = 1558  # UINT32 deceleration
        RAMP_N_MAX = 1554  # UINT16 max rpm
        FLT_INFO = 15362  # 22 registers, code for error
        FLT_MEM_RESET = 15114  # UINT16 reset fault memory
        FLT_MEM_DEL = 15112  # UINT16 delete fault memory

    class Mode(IntEnum):
        """
        ILS2T device modes
        """
        PTP = 3  # point to point
        JOG = 1

    class ActionsPtp(IntEnum):
        """
        Allowed actions in the point to point mode (`ILS2T.Mode.PTP`).
        """
        ABSOLUTE_POSITION = 0
        RELATIVE_POSITION_TARGET = 1
        RELATIVE_POSITION_MOTOR = 2

    ACTION_JOG_VALUE = 0
    """
    The single action value for `ILS2T.Mode.JOG`
    """

    # Note: don't use IntFlag here - it allows other then enumerated values
    class Ref16Jog(Flag):
        """
        Allowed values for ILS2T ref_16 register (the shown values are the integer
        representation of the bits), all in Jog mode = 1
        """
        NONE = 0
        POS = 1
        NEG = 2
        FAST = 4
        # allowed combinations
        POS_FAST = POS | FAST
        NEG_FAST = NEG | FAST

    class State(IntEnum):
        """
        State machine status values
        """
        QUICKSTOP = 7
        READY = 4
        ON = 6

    DEFAULT_IO_SCANNING_CONTROL_VALUES = {
        'action': ActionsPtp.RELATIVE_POSITION_MOTOR.value,
        'mode': Mode.PTP.value,
        'disable_driver_di': 0,
        'enable_driver_en': 0,
        'quick_stop_qs': 0,
        'fault_reset_fr': 0,
        'execute_stop_sh': 0,
        'reset_stop_ch': 0,
        'continue_after_stop_cu': 0,
        'ref_16': ILS2TConfig.rpm_max_init,
        'ref_32': 0,
    }
    """
    Default IO Scanning control mode values
    """

    def __init__(self, com, dev_config=None) -> None:
        """
        Constructor for ILS2T.

        :param com: object to use as communication protocol.
        """

        # Call superclass constructor
        super().__init__(com, dev_config)

        # toggle reminder bit
        self._mode_toggle_mt = 0
        self.flt_list = []

    @staticmethod
    def default_com_cls():
        return ILS2TModbusTcpCommunication

    @staticmethod
    def config_cls():
        return ILS2TConfig

    def start(self) -> None:
        """
        Start this device.
        """

        logging.info("Starting device " + str(self))

        try:
            # try opening the port
            self.com.open()
        except ModbusTcpConnectionFailedException as exc:
            logging.error(exc)
            raise

        # writing 1 to register ACCESS_ENABLE allows to use the IO scanning mode.
        #  This is not documented in the manual!
        self.com.write_registers(self.RegAddr.ACCESS_ENABLE.value,
                                 [0, 1])

        # set maximum RPM from init config
        self.set_max_rpm(self.config.rpm_max_init)

    def stop(self) -> None:
        """
        Stop this device. Disables the motor (applies brake), disables access and
        closes the communication protocol.
        """

        logging.info("Stopping device " + str(self))
        self.disable()
        self.com.write_registers(self.RegAddr.ACCESS_ENABLE.value,
                                 [0, 0])
        self.com.close()

    def get_status(self) -> Dict[str, int]:
        """
        Perform an IO Scanning read and return the status of the motor.

        :return: dict with status information.
        """

        registers = self.com \
            .read_holding_registers(self.RegAddr.IO_SCANNING.value, 8)
        return self._decode_status_registers(registers)

    def do_ioscanning_write(self, **kwargs: int) -> None:
        """
        Perform a write operation using IO Scanning mode.

        :param kwargs:
            Keyword-argument list with options to send, remaining are taken
            from the defaults.
        """

        self._toggle()
        values = self._generate_control_registers(**kwargs)
        self.com.write_registers(self.RegAddr.IO_SCANNING.value,
                                 values)

    def _generate_control_registers(self, **kwargs: int) -> List[int]:
        """
        Generates the control registers for the IO scanning mode.
        It is necessary to write all 64 bit at the same time, so a list of 4 registers
        is generated.

        :param kwargs: Keyword-argument list with options different than the defaults.
        :return: List of registers for the IO scanning mode.
        """

        cleaned_io_scanning_mode = self._clean_ioscanning_mode_values(kwargs)

        action_bits = '{0:03b}'.format(cleaned_io_scanning_mode['action'])
        mode_bits = '{0:04b}'.format(cleaned_io_scanning_mode['mode'])
        builder = BinaryPayloadBuilder(byteorder=Endian.Big, wordorder=Endian.Big)

        # add the first byte: Drive control
        builder.add_bits(
            [
                cleaned_io_scanning_mode['disable_driver_di'],
                cleaned_io_scanning_mode['enable_driver_en'],
                cleaned_io_scanning_mode['quick_stop_qs'],
                cleaned_io_scanning_mode['fault_reset_fr'],
                0,  # has to be 0 per default, no meaning
                cleaned_io_scanning_mode['execute_stop_sh'],
                cleaned_io_scanning_mode['reset_stop_ch'],
                cleaned_io_scanning_mode['continue_after_stop_cu'],
            ]
        )

        # add the second byte: Mode control
        builder.add_bits(
            [
                int(mode_bits[3]),
                int(mode_bits[2]),
                int(mode_bits[1]),
                int(mode_bits[0]),
                int(action_bits[2]),
                int(action_bits[1]),
                int(action_bits[0]),
                self._mode_toggle_mt,
            ]
        )

        # add the third and fourth byte:
        # Ref_16 (either JOG direction/speed, or RPM...)
        builder.add_16bit_uint(cleaned_io_scanning_mode['ref_16'])

        # add 4 bytes Ref_32, Target position
        builder.add_32bit_int(cleaned_io_scanning_mode['ref_32'])

        return builder.to_registers()

    def _clean_ioscanning_mode_values(self, io_scanning_values: Dict[str, int]) \
            -> Dict[str, int]:
        """
        Checks if the constructed mode is valid.

        :param io_scanning_values: Dictionary with register values to check
        :return: Dictionary with cleaned register values
        :raises ValueError: if `io_scanning_values` has unrecognized keys
        :raises IoScanningModeValueError: if either `'mode'` or either of corresponding
            `'action'`, `'ref_16'`, or `'ref_32'` keys of `io_scanning_values` has
            an invalid value.
        """

        # check if there are too much keys that are not recognized
        io_scanning_keys = set(io_scanning_values.keys())
        all_keys = set(self.DEFAULT_IO_SCANNING_CONTROL_VALUES.keys())
        superfluous_keys = io_scanning_keys.difference(all_keys)
        if superfluous_keys:
            raise ValueError("Unrecognized mode keys: {}".format(
                list(superfluous_keys)))

        # fill up io_scanning_values with defaults, if they are not set
        for mode_key, default_value in self.DEFAULT_IO_SCANNING_CONTROL_VALUES.items():
            if mode_key not in io_scanning_values:
                io_scanning_values[mode_key] = default_value

        # perform checks depending on mode
        # JOG mode
        if io_scanning_values['mode'] == self.Mode.JOG:

            if not io_scanning_values['action'] == self.ACTION_JOG_VALUE:
                raise IoScanningModeValueError(
                    'Wrong action: {}'.format(io_scanning_values['action'])
                )

            try:
                self.Ref16Jog(io_scanning_values['ref_16'])
            except ValueError:
                raise IoScanningModeValueError(
                    'Wrong value in ref_16 ({})'.format(io_scanning_values['ref_16'])
                )

            if not io_scanning_values['ref_32'] == 0:
                raise IoScanningModeValueError(
                    'Wrong value in ref_32 ({})'.format(io_scanning_values['ref_32'])
                )

            return io_scanning_values

        # PTP mode
        if io_scanning_values['mode'] == self.Mode.PTP:

            try:
                self.ActionsPtp(io_scanning_values['action'])
            except ValueError:
                raise IoScanningModeValueError(
                    'Wrong action: {}'.format(io_scanning_values['action'])
                )

            if not self._is_valid_rpm(io_scanning_values['ref_16']):
                raise IoScanningModeValueError(
                    'Wrong value in ref_16 ({})'.format(io_scanning_values['ref_16'])
                )

            if not self._is_int32(io_scanning_values['ref_32']):
                raise IoScanningModeValueError(
                    'Wrong value in ref_32 ({})'.format(io_scanning_values['ref_32'])
                )

            return io_scanning_values

        # default
        raise IoScanningModeValueError(
            'Wrong mode: {}'.format(io_scanning_values['mode'])
        )

    def _is_valid_rpm(self, num: int) -> bool:
        """
        Checks whether `num` is a valid RPM value.

        :param num: RPM value to check
        :return: `True` if `num` is a valid RPM value, `False` otherwise
        """

        return isinstance(num, int) and 0 < num <= self.config.rpm_max_init

    @classmethod
    def _is_int32(cls, num: int) -> bool:
        """
        Checks whether a number fits in a signed 32-bit integer.

        :param num: is the number to check.
        :return: check result.
        """

        return isinstance(num, int) and cls.RegDatatype.INT32.is_in_range(num)

    @staticmethod
    def _decode_status_registers(registers: List[int]) -> Dict[str, int]:
        """
        Decodes the the status of the stepper drive, derived from IOscanning.

        :param registers: List of 8 registers (6922-6930)
        :return: dict
        """

        decoder = BinaryPayloadDecoder.fromRegisters(registers, byteorder=Endian.Big)
        decoded = {
            'drive_control': decoder.decode_bits(),
            'mode_control': decoder.decode_bits(),
            'ref_16': decoder.decode_16bit_int(),
            'ref_32': decoder.decode_32bit_int(),
            'drive_status_1': decoder.decode_bits(),
            'drive_status_2': decoder.decode_bits(),
            'mode_status': decoder.decode_bits(),
            'drive_input': decoder.decode_bits(),
            'action_word_1': decoder.decode_bits(),
            'action_word_2': decoder.decode_bits(),
            'special_function_1': decoder.decode_bits(),
            'special_function_2': decoder.decode_bits(),
        }

        return {
            'mode': BitArray(decoded['mode_status'][3::-1]).int,
            'action': BitArray(decoded['mode_control'][6:3:-1]).int,
            'ref_16': decoded['ref_16'],
            'ref_32': decoded['ref_32'],
            'state': BitArray(decoded['drive_status_2'][3::-1]).int,
            'fault': decoded['drive_status_2'][6],
            'warn': decoded['drive_status_2'][7],
            'halt': decoded['drive_status_1'][0],
            'motion_zero': decoded['action_word_2'][6],
            'turning_positive': decoded['action_word_2'][7],
            'turning_negative': decoded['action_word_1'][0],
        }

    def _toggle(self) -> None:
        """
        To activate a command it is necessary to toggle the MT bit first.
        """

        self._mode_toggle_mt = (0 if self._mode_toggle_mt else 1)

    def relative_step(self, steps: int) -> None:
        """
        Turn the motor the relative amount of steps. This function does not enable or
        disable the motor automatically.
        positive numbers -> CW
        negative numbers -> CCW

        :param steps: Number of steps to turn the motor.
        """

        if not abs(steps) < 2_147_483_647:
            logging.warning('number of steps is too big: {}'.format(steps))

        logging.info('Perform number of steps: {}'.format(steps))

        self.do_ioscanning_write(
            enable_driver_en=1,
            mode=self.Mode.PTP.value,
            action=self.ActionsPtp.RELATIVE_POSITION_MOTOR.value,
            ref_32=steps
        )

    def absolute_position(self, position: int) -> None:
        """
        Turn the motor until it reaches the absolute position.
        This function does not enable or disable the motor automatically.

        :param position: absolute position of motor in user defined steps.
        """

        if not abs(position) < 2_147_483_647:
            logging.warning('position is out of range: {}'.format(position))

        logging.info('Absolute position: {}'.format(position))

        self.do_ioscanning_write(
            enable_driver_en=1,
            mode=self.Mode.PTP.value,
            action=self.ActionsPtp.ABSOLUTE_POSITION.value,
            ref_32=position
        )

    def relative_step_and_wait(self, steps: int) -> None:
        """
        Enable motor, perform relative steps and wait until done, disable.

        :param steps: Number of steps.
        """

        logging.info('Motor steps requested: {}'.format(steps))

        position_before = self.get_position()

        self.enable()
        sleep(1)
        self.relative_step(steps)
        sleep(2)

        while True:
            if self.get_status()['motion_zero']:
                self.disable()
                break

        # check if steps were made
        position_after = self.get_position()
        if position_before + steps != position_after:
            flt_dict = self.get_error_code()
            self.flt_list.append(flt_dict)
            if 'empty' in flt_dict[0].keys():
                logging.warning('no error in drive, '
                                'something different must have gone wrong')
                logging.warning(
                    'The position does not align with the requested step '
                    'number. Before: {0}, after: {1}, requested: {2}, '
                    'real difference: {3}'.format(
                        position_before,
                        position_after,
                        steps,
                        position_after - position_before,
                    )
                )
            else:
                # Despite drive error/malfunction don't break the code/experiment
                # execution by raising an error; continuing as nothing happened.
                logging.critical('error in drive, drive is know maybe locked')
                logging.critical(
                    'The position does not align with the requested step '
                    'number. Before: {0}, after: {1}, requested: {2}, '
                    'real difference: {3}'.format(
                        position_before,
                        position_after,
                        steps,
                        position_after - position_before,
                    )
                )

    def absolute_position_and_wait(self, position: int) -> None:
        """
        Enable motor, perform absolute position and wait until done, disable.

        :param position: absolute position of motor in user defined steps.
        """

        logging.info('absolute position requested: {}'.format(position))

        position_before = self.get_position()

        self.enable()
        sleep(1)
        self.absolute_position(position)
        sleep(2)

        while True:
            if self.get_status()['motion_zero']:
                self.disable()
                break

        # check if steps were made
        position_after = self.get_position()
        if position != position_after:
            flt_dict = self.get_error_code()
            self.flt_list.append(flt_dict)
            if 'empty' in flt_dict[0].keys():
                logging.warning('no error in drive, '
                                'something different must have gone wrong')
                logging.warning(
                    'The position does not align with the requested absolute position. '
                    'Before: {0}, after: {1}, requested: {2}'
                    .format(
                        position_before,
                        position_after,
                        position,
                    )
                )
            else:
                logging.critical('error in drive, drive is know maybe locked')
                logging.critical(
                    'The position does not align with the requested absolute position. '
                    'Before: {0}, after: {1}, requested: {2}'
                    .format(
                        position_before,
                        position_after,
                        position,
                    )
                )

    def disable(self) -> None:
        """
        Disable the driver of the stepper motor and enable the brake.
        """

        if self.get_status()['motion_zero']:
            logging.info('Disable motor, brake.')
            self.do_ioscanning_write(enable_driver_en=0, disable_driver_di=1)
        else:
            logging.warning('Cannot disable motor, still running!')

    def enable(self) -> None:
        """
        Enable the driver of the stepper motor and disable the brake.
        """

        self.do_ioscanning_write(enable_driver_en=1, disable_driver_di=0)
        logging.info('Enable motor, disable brake.')

    def get_position(self) -> int:
        """
        Read the position of the drive and store into status.

        :return: Position step value
        """

        value = self.com.read_input_registers(
            self.RegAddr.POSITION.value, 2)
        return self._decode_32bit(value, True)

    def get_temperature(self) -> int:
        """
        Read the temperature of the motor.

        :return: Temperature in degrees Celsius.
        """

        value = self.com.read_input_registers(
            self.RegAddr.TEMP.value, 2)
        return self._decode_32bit(value, True)

    def get_dc_volt(self) -> float:
        """
        Read the DC supply voltage of the motor.

        :return: DC input voltage.
        """

        value = self.com.read_input_registers(
            self.RegAddr.VOLT.value, 2)
        return self._decode_32bit(value, True) / 10

    @staticmethod
    def _decode_32bit(registers: List[int], signed: bool = True) -> int:
        """
        Decode two 16-bit ModBus registers to a 32-bit integer.

        :param registers: list of two register values
        :param signed: True, if register containes a signed value
        :return: integer representation of the 32-bit register
        """

        decoder = BinaryPayloadDecoder.fromRegisters(registers, byteorder=Endian.Big)
        if signed:
            return decoder.decode_32bit_int()
        else:
            return decoder.decode_32bit_uint()

    def user_steps(self, steps: int = 16384, revolutions: int = 1) -> None:
        """
        Define steps per revolution.
        Default is 16384 steps per revolution.
        Maximum precision is 32768 steps per revolution.

        :param steps: number of steps in `revolutions`.
        :param revolutions: number of revolutions corresponding to `steps`.
        """

        if not self._is_int32(revolutions):
            err_msg = 'Wrong scaling factor: revolutions = {}'.format(revolutions)
            logging.error(err_msg)
            raise ScalingFactorValueError(err_msg)

        if not self._is_int32(steps):
            err_msg = 'Wrong scaling factor: steps = {}'.format(steps)
            logging.error(err_msg)
            raise ScalingFactorValueError(err_msg)

        builder = BinaryPayloadBuilder(byteorder=Endian.Big, wordorder=Endian.Big)
        builder.add_32bit_int(steps)
        builder.add_32bit_int(revolutions)
        values = builder.to_registers()
        self.com.write_registers(self.RegAddr.SCALE.value, values)

    def quickstop(self) -> None:
        """
        Stops the motor with high deceleration rate and falls into error state. Reset
        with `reset_error` to recover into normal state.
        """

        logging.warning('Motor QUICK STOP.')
        self.do_ioscanning_write(quick_stop_qs=1)

    def reset_error(self) -> None:
        """
        Resets the motor into normal state after quick stop or another error occured.
        """

        logging.info('Reset motor after fault or quick stop.')
        self.do_ioscanning_write(fault_reset_fr=1)

    def jog_run(self, direction: bool = True, fast: bool = False) -> None:
        """
        Slowly turn the motor in positive direction.
        """

        status = self.get_status()

        if status['mode'] != self.Mode.JOG and not status['motion_zero']:
            logging.error('Motor is not in Jog mode or standstill, abort.')
            return

        if status['state'] != self.State.ON:
            # need to enable first
            logging.error('Motor is not enabled or in error state. Try .enable()')
            return

        ref_16 = self.Ref16Jog.NONE

        if direction:
            ref_16 = ref_16 | self.Ref16Jog.POS
            logging.info('Jog mode in positive direction enabled.')
        else:
            ref_16 = ref_16 | self.Ref16Jog.NEG
            logging.info('Jog mode in negative direction enabled.')

        if fast:
            ref_16 = ref_16 | self.Ref16Jog.FAST

        self.do_ioscanning_write(
            mode=self.Mode.JOG.value,
            action=self.ACTION_JOG_VALUE,
            enable_driver_en=1,
            ref_16=ref_16.value,
        )

    def jog_stop(self) -> None:
        """
        Stop turning the motor in Jog mode.
        """

        logging.info('Stop in Jog mode.')

        self.do_ioscanning_write(
            mode=self.Mode.JOG.value,
            action=self.ACTION_JOG_VALUE,
            enable_driver_en=1,
            ref_16=0,
        )

    def set_jog_speed(self, slow: int = 60, fast: int = 180) -> None:
        """
        Set the speed for jog mode. Default values correspond to startup values of
        the motor.

        :param slow: RPM for slow jog mode.
        :param fast: RPM for fast jog mode.
        """

        logging.info(
            'Setting Jog RPM. Slow = {0} RPM, Fast = {1} RPM.'.format(
                slow, fast
            )
        )
        self.com.write_registers(self.RegAddr.JOGN_SLOW.value,
                                 [0, slow])
        self.com.write_registers(self.RegAddr.JOGN_FAST.value,
                                 [0, fast])

    def get_error_code(self) -> Dict[int, Dict[str, Any]]:
        """
        Read all messages in fault memory.
        Will read the full error message and return the decoded values.
        At the end the fault memory of the motor will be deleted.
        In addition, reset_error is called to re-enable the motor for operation.

        :return: Dictionary with all information
        """

        ret_dict = {}
        self.com.write_registers(self.RegAddr.FLT_MEM_RESET.value,
                                 [0, 1])
        for i in range(10):
            registers = self.com.read_input_registers(
                self.RegAddr.FLT_INFO.value, 22)
            decoder = BinaryPayloadDecoder.fromRegisters(registers,
                                                         byteorder=Endian.Big)
            decoded = {
                'ignored0': decoder.skip_bytes(2),
                'error_code': decoder.decode_16bit_uint(),
                'ignored1': decoder.skip_bytes(2),
                'error_class': decoder.decode_16bit_uint(),
                'error_time': decoder.decode_32bit_uint(),
                'ignored2': decoder.skip_bytes(2),
                'error_addition': decoder.decode_16bit_uint(),
                'ignored3': decoder.skip_bytes(2),
                'error_no_cycle': decoder.decode_16bit_uint(),
                'ignored4': decoder.skip_bytes(2),
                'error_after_enable': decoder.decode_16bit_uint(),
                'ignored5': decoder.skip_bytes(2),
                'error_voltage_dc': decoder.decode_16bit_uint(),
                'ignored6': decoder.skip_bytes(2),
                'error_rpm': decoder.decode_16bit_int(),
                'ignored7': decoder.skip_bytes(2),
                'error_current': decoder.decode_16bit_uint(),
                'ignored8': decoder.skip_bytes(2),
                'error_amplifier_temperature': decoder.decode_16bit_int(),
                'ignored9': decoder.skip_bytes(2),
                'error_device_temperature': decoder.decode_16bit_int(),
            }
            flt_dict = {
                'error_code': hex(decoded['error_code'])[2:],
                'error_class': decoded['error_class'],
                'error_time': timedelta(seconds=decoded['error_time']),
                'error_addition': decoded['error_addition'],
                'error_no_cycle': decoded['error_no_cycle'],
                'error_after_enable': timedelta(seconds=decoded['error_after_enable']),
                'error_voltage_dc': decoded['error_voltage_dc'] / 10,
                'error_rpm': decoded['error_rpm'],
                'error_current': decoded['error_current'] / 100,
                'error_amplifier_temperature': decoded['error_amplifier_temperature'],
                'error_device_temperature': decoded['error_device_temperature'],
            }
            ret_dict[i] = flt_dict
            if flt_dict['error_code'] == '0':
                flt_dict = {'empty': None}
                ret_dict = {i: flt_dict}
                break
        self.com.write_registers(self.RegAddr.FLT_MEM_DEL.value,
                                 [0, 1])
        self.reset_error()
        return ret_dict

    def set_max_rpm(self, rpm: int) -> None:
        """
        Set the maximum RPM.

        :param rpm: revolution per minute ( 0 < rpm <= RPM_MAX)
        :raises ILS2TException: if RPM is out of range
        """

        if self._is_valid_rpm(rpm):
            self.DEFAULT_IO_SCANNING_CONTROL_VALUES['ref_16'] = rpm
            self.com.write_registers(
                self.RegAddr.RAMP_N_MAX.value, [0, rpm])
        else:
            raise ILS2TException('RPM out of range: {rpm} not in (0, {rpm_max}'.format(
                rpm=rpm,
                rpm_max=self.config.rpm_max_init,
            ))

    def set_ramp_type(self, ramp_type: int = -1) -> None:
        """
        Set the ramp type. There are two options available:
            0:  linear ramp
            -1: motor optimized ramp

        :param ramp_type: 0: linear ramp | -1: motor optimized ramp
        """

        self.com.write_registers(
            self.RegAddr.RAMP_TYPE.value, [0, ramp_type]
        )

    def set_max_acceleration(self, rpm_minute: int) -> None:
        """
        Set the maximum acceleration of the motor.

        :param rpm_minute: revolution per minute per minute
        """

        builder = BinaryPayloadBuilder(byteorder=Endian.Big, wordorder=Endian.Big)
        builder.add_32bit_uint(rpm_minute)
        values = builder.to_registers()
        self.com.write_registers(self.RegAddr.RAMP_ACC.value, values)

    def set_max_deceleration(self, rpm_minute: int) -> None:
        """
        Set the maximum deceleration of the motor.

        :param rpm_minute: revolution per minute per minute
        """

        builder = BinaryPayloadBuilder(byteorder=Endian.Big, wordorder=Endian.Big)
        builder.add_32bit_uint(rpm_minute)
        values = builder.to_registers()
        self.com.write_registers(self.RegAddr.RAMP_DECEL.value,
                                 values)



#  Copyright (c) 2019 ETH Zurich, SIS ID and HVL D-ITET
#
"""
Additional Python typing module utilities
"""

from typeguard import origin_type_checkers


def is_type_hint(type_):
    """
    Check if class is a generic type, for example `Union` or `List[int]`

    :param type_: type to check
    """
    if not hasattr(type_, '__module__'):
        return False

    # check if the class inherits from any `typing` class
    if type_.__module__ != 'typing':
        if not any(t.__module__ == 'typing' for t in type_.mro()):
            return False

    origin = getattr(type_, '__origin__', type_)  # Note: List.__origin__ == list

    return origin in origin_type_checkers


def is_generic(type_):
    """
    Check if class is a user-defined generic type, for example `Union[int, float]` but
    not `List`.

    :param type_: type to check
    """
    return is_type_hint(type_) and bool(getattr(type_, '__args__', ()))


def check_generic_type(value, type_, name="instance"):
    """
    Check if `value` is of a generic type `type_`. Raises `TypeError` if it's not.

    :param name: name to report in case of an error
    :param value: value to check
    :param type_: generic type to check against
    """
    if not hasattr(type_, '__origin__') or type_.__origin__ not in origin_type_checkers:
        raise ValueError('Type {} is not a generic type.'.format(type_))
    origin_type_checkers[type_.__origin__](name, value, type_, None)



"""
Communication protocol for serial ports. Makes use of the `pySerial
<https://pythonhosted.org/pyserial/index.html>`_ library.
"""

# Note: PyCharm does not recognize the dependency correctly, it is added as pyserial.
import serial

from .base import CommunicationProtocol
from ..configuration import configdataclass
from ..utils.enum import ValueEnum, unique


@configdataclass
class SerialCommunicationConfig:
    """
    Configuration dataclass for :class:`SerialCommunication`.
    """

    @unique
    class Parity(ValueEnum):
        """
        Serial communication parity.
        """
        EVEN = serial.PARITY_EVEN
        MARK = serial.PARITY_MARK
        NAMES = serial.PARITY_NAMES
        NONE = serial.PARITY_NONE
        ODD = serial.PARITY_ODD
        SPACE = serial.PARITY_SPACE

    @unique
    class Stopbits(ValueEnum):
        """
        Serial communication stopbits.
        """
        ONE = serial.STOPBITS_ONE
        ONE_POINT_FIVE = serial.STOPBITS_ONE_POINT_FIVE
        TWO = serial.STOPBITS_TWO

    @unique
    class Bytesize(ValueEnum):
        """
        Serial communication bytesize.
        """
        FIVEBITS = serial.FIVEBITS
        SIXBITS = serial.SIXBITS
        SEVENBITS = serial.SEVENBITS
        EIGHTBITS = serial.EIGHTBITS

    #: Port is a string referring to a COM-port (e.g. ``'COM3'``) or a URL.
    #: The full list of capabilities is found `on the pyserial documentation
    #: <https://pythonhosted.org/pyserial/url_handlers.html>`_.
    port: str

    #: Baudrate of the serial port
    baudrate: int

    #: Parity to be used for the connection.
    parity: (str, Parity)

    #: Stopbits setting, can be 1, 1.5 or 2.
    stopbits: (int, float, Stopbits)

    #: Size of a byte, 5 to 8
    bytesize: (int, Bytesize)

    #: The terminator character. Typically this is ``b'\r\n'`` or ``b'\n'``, but can
    #: also be ``b'\r'`` or other combinations.
    terminator: bytes = b'\r\n'

    #: Timeout in seconds for the serial port
    timeout: (int, float) = 2

    def clean_values(self):
        if not isinstance(self.parity, self.Parity):
            self.force_value('parity', self.Parity(self.parity))

        if not isinstance(self.stopbits, self.Stopbits):
            self.force_value('stopbits', self.Stopbits(self.stopbits))

        if not isinstance(self.bytesize, self.Bytesize):
            self.force_value('bytesize', self.Bytesize(self.bytesize))


class SerialCommunication(CommunicationProtocol):
    """
    Implements the Communication Protocol for serial ports.
    """

    ENCODING = 'utf-8'
    UNICODE_HANDLING = 'replace'

    def __init__(self, configuration):
        """
        Constructor for SerialCommunication.
        """

        super().__init__(configuration)

        # create the serial port specified in the configuration
        with serial.serial_for_url(self.config.port, do_not_open=True) as self.ser:
            self.ser.baudrate = self.config.baudrate
            self.ser.parity = self.config.parity.value
            self.ser.stopbits = self.config.stopbits.value
            self.ser.bytesize = self.config.bytesize.value
            self.ser.timeout = self.config.timeout

    @staticmethod
    def config_cls():
        return SerialCommunicationConfig

    def open(self):
        """
        Open the serial connection.
        """

        # open the port
        with self.access_lock:
            self.ser.open()

    def close(self):
        """
        Close the serial connection.
        """

        # close the port
        with self.access_lock:
            self.ser.close()

    def write_text(self, text: str):
        """
        Write text to the serial port. The text is encoded and terminated by
        the configured terminator.

        :param text: Text to send to the port.
        """

        with self.access_lock:
            self.ser.write(text.encode(self.ENCODING, self.UNICODE_HANDLING)
                           + self.config.terminator)

    def read_text(self) -> str:
        """
        Read one line of text from the serial port. The input buffer may
        hold additional data afterwards, since only one line is read.

        :return: String read from the serial port.
        """

        with self.access_lock:
            decoded = self.ser.readline().decode(self.ENCODING)

        return decoded



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""The setup script."""

from setuptools import setup, find_packages

with open('README.rst') as readme_file:
    readme = readme_file.read()

with open('HISTORY.rst') as history_file:
    history = history_file.read()

requirements = [
    'pyserial>=3.4',
    'LJMPython>=1.19.0',
    'pymodbus==2.1.0',
    'IPy>=0.83',
    'bitstring>=3.1.5',
    'pyvisa>=1.9.1',
    'pyvisa-py>=0.3.1',
    'typeguard>=2.3.0',
    'aenum>=2.1.2',
    'opcua>=0.98.6',
    'cryptography>=2.6.1',  # optional dependency of the opcua package
]

dependency_links = [
    'https://labjack.com/sites/default/files/software/Python_LJM_2018_10_19.zip'
    '#egg=LJMPython-1.19.0',
]

setup_requirements = ['pytest-runner', ]

test_requirements = ['pytest', ]

setup(
    author="Mikołaj Rybiński, David Graber, Henrik Menne",
    author_email='mikolaj.rybinski@id.ethz.ch, graber@eeh.ee.ethz.ch, '
                 'henrik.menne@eeh.ee.ethz.ch',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Natural Language :: English',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.7',
    ],
    description="Python common code base to control devices used in Christian Franck's High Voltage Lab (HVL), D-ITET, ETH",
    entry_points={
    },
    install_requires=requirements,
    dependency_links=dependency_links,
    license="GNU General Public License v3",
    long_description=readme + '\n\n' + history,
    include_package_data=True,
    keywords='hvl_ccb',
    name='hvl_ccb',
    packages=find_packages(),
    setup_requires=setup_requirements,
    test_suite='tests',
    tests_require=test_requirements,
    url='https://gitlab.ethz.ch/hvl_priv/hvl_ccb',
    version='0.3.0',
    zip_safe=False,
)
